//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "downloadqueue.h"
#include "updownclient.h"
#include "partfile.h"
#include "ed2klink.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


CDownloadQueue::CDownloadQueue(CPreferences* in_prefs,CSharedFileList* in_sharedfilelist){
	app_prefs = in_prefs;
	sharedfilelist = in_sharedfilelist;
	filesrdy = 0;
	datarate = 0;
	cur_udpserver = 0;
	lastfile = 0;
	lastudpsearchtime = 0;
	lastudpstattime = 0;
	udcounter = 0;
	m_nDownDataRateMSOverhead = 0;
	m_nDownDatarateOverhead = 0;
	m_nDownDataOverheadSourceExchange = 0;
	m_nDownDataOverheadFileRequest = 0;
	m_nDownDataOverheadOther = 0;
	m_nDownDataOverheadServer = 0;
	m_nDownDataOverheadSourceExchangePackets = 0;
	m_nDownDataOverheadFileRequestPackets = 0;
	m_nDownDataOverheadOtherPackets = 0;
	m_nDownDataOverheadServerPackets = 0;
}

void CDownloadQueue::UpdateDisplayedInfo(boolean force) {
    DWORD curTick = GetTickCount();

    if(force || curTick-m_lastRefreshedDLDisplay > MINWAIT_BEFORE_DLDISPLAY_WINDOWUPDATE+(uint32)(rand()/(RAND_MAX/1000))) {
        theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
        m_lastRefreshedDLDisplay = curTick;
    }
}

void CDownloadQueue::AddPartFilesToShare()
{
	EMULE_TRY
    for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		if (cur_file->GetStatus(true) == PS_READY)
			sharedfilelist->SafeAddKFile(cur_file,true);
	}
    EMULE_CATCH
}

void CDownloadQueue::CompDownDatarateOverhead(){
    EMULE_TRY
	m_AvarageDDRO_list.AddTail(m_nDownDataRateMSOverhead);
	if (m_AvarageDDRO_list.GetCount() > 150)
		m_AvarageDDRO_list.RemoveAt(m_AvarageDDRO_list.GetHeadPosition());

	m_nDownDatarateOverhead = 0;
	m_nDownDataRateMSOverhead = 0;
	for (POSITION pos = m_AvarageDDRO_list.GetHeadPosition();pos != 0;m_AvarageDDRO_list.GetNext(pos))
		m_nDownDatarateOverhead += m_AvarageDDRO_list.GetAt(pos);

	if(m_AvarageDDRO_list.GetCount() > 10)
		m_nDownDatarateOverhead = 10*m_nDownDatarateOverhead/m_AvarageDDRO_list.GetCount();
	else
		m_nDownDatarateOverhead = 0;
	return;
    EMULE_CATCH
}



void CDownloadQueue::Init()
{
    EMULE_TRY
	//////////////////////////////////////////////////
	//for saving/loading sources mr fry
	
	//this indicates, if a permanent open CStoredSources-object for each
	//part-file should be used...
	//means, additons of sources and loading sources are made to/from this
	//open CStoredSources-object and saved to disk only on closing of eMule
	//
	//otherwise on each saving-/loading- session the *.src-file on disk is
	//opened before the save-/load- operation and closed again afterwards
	//
	//I think the difference is:
	//set to true: more memory is used (I think, it took about 20 kB to save 300 sources to my hard-drive)
	//set to false: more cpu-power is used
	m_bKeepStoredSourcesOpened	= true;
		
	if(app_prefs->SLSEnable())
		SetSourcesLoading(true);
	else
		SetSourcesLoading(false);
	//////////////////////////////////////////////////

	// find all part files, read & hash them if needed and store into a list
	// start changed by InterCeptor (multiple tempdirs) 19.11.02
	uint16 count = 0;
	app_prefs->tempdir_list.AddHead(app_prefs->GetTempDir());
	for (POSITION pos = app_prefs->tempdir_list.GetHeadPosition();pos != 0;app_prefs->tempdir_list.GetNext(pos))
	{
		CFileFind ff;
		CString searchPath(app_prefs->tempdir_list.GetAt(pos).GetBuffer());
		searchPath += "\\*.part.met";

		//check all part.met files
		bool end = !ff.FindFile(searchPath,0);
		while (!end)
		{
			end = !ff.FindNextFile();
			if (ff.IsDirectory())
				continue;
			CPartFile* toadd = new CPartFile();
			if (toadd->LoadPartFile(app_prefs->tempdir_list.GetAt(pos).GetBuffer(),ff.GetFileName().GetBuffer()))
			{
				count++;
				// Added by Tarod -check if is an A4AF auto file, correct place?
				if (memcmp(toadd->GetFileHash(), theApp.glob_prefs->GetA4AFHash(), 16) == 0)
					toadd->SetA4AFAuto(true);
				filelist.AddTail(toadd); // to downloadqueue
				if (toadd->GetStatus(true) == PS_READY)
					sharedfilelist->SafeAddKFile(toadd); // part files are always shared files
				theApp.emuledlg->transferwnd.downloadlistctrl.AddFile(toadd);// show in downloadwindow
				//////////////////////////////////////////////////
				//for loading sources mr fry
#ifdef OLD_SLS_ENABLED
				if(toadd->GetStatus() < 4 && IsSourcesLoadingEnabled())
				{
							
				
					//load source for the part-file
					toadd->LoadAndAddSources(KeepStoredSourcesOpen(),false,toadd->GetLoadSourcesAtOnceLimit(),true,false,false);	//to restore sources from file (by Mr.Fry)
					toadd->SetLoadedSourcesCompletely(false);

					//do load sources for part-files with different time-offsets
					//means: do not load the sources of all part-files at one time
					//(offset: (No. of part-file) multiplied by (1 second))
					toadd->SetLoadSourcesSlowTimeInterval((toadd->GetLoadSourcesSlowTimeInterval() + (count * 1000)));
				}
#endif //OLD_SLS_ENABLED
				//////////////////////////////////////////////////
			}
			else
				delete toadd;
		}
		ff.Close();

		//try recovering any part.met files
		searchPath += ".backup";
		end = !ff.FindFile(searchPath, 0);
		while (!end){
			end = !ff.FindNextFile();
			if (ff.IsDirectory())
				continue;
			CPartFile* toadd = new CPartFile();
			if (toadd->LoadPartFile(app_prefs->tempdir_list.GetAt(pos).GetBuffer(),ff.GetFileName().GetBuffer()))
			{
				toadd->SavePartFile(); // resave backup
				count++;
				filelist.AddTail(toadd);			// to downloadqueue
				if (toadd->GetStatus(true) == PS_READY)
					sharedfilelist->SafeAddKFile(toadd); // part files are always shared files
				theApp.emuledlg->transferwnd.downloadlistctrl.AddFile(toadd);// show in downloadwindow

				theApp.emuledlg->AddLogLine(false, GetResString(IDS_RECOVERED_PARTMET), toadd->GetFileName());
			}
			else
			{
				delete toadd;
			}
		}
		ff.Close();
	}
	app_prefs->tempdir_list.RemoveHead();
	// end added by InterCeptor (multiple tempdirs) 19.11.02
	if(count == 0) {
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_NOPARTSFOUND));
	} else {
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_FOUNDPARTS),count);
		SortByPriority();
	}

	//--- xrmb:keepPartFileStats ---
	m_lastPartFileStatsSave=GetTickCount();
	//--- :xrmb ---
    EMULE_CATCH
}

CDownloadQueue::~CDownloadQueue(){
    EMULE_TRY
	//--- xrmb:keepPartFileStats ---
	SaveAllPartFileStats();
	//--- :xrmb ---

	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		filelist.GetAt(pos)->m_hpartfile.Flush() ; // Flush before close app [Tarod]
		filelist.GetAt(pos)->SavePartFile();
		delete filelist.GetAt(pos);
	}
    EMULE_CATCH
}

void CDownloadQueue::AddSearchToDownload(CSearchFile* toadd){
    EMULE_TRY
	if (IsFileExisting(toadd->GetFileHash()))
		return;
	CPartFile* newfile = new CPartFile(toadd);
	if (newfile->GetStatus() == PS_ERROR){
		delete newfile;
		return;
	}
	AddDownload(newfile);
    EMULE_CATCH
}

void CDownloadQueue::AddSearchToDownload(CString link){
    EMULE_TRY
	CPartFile* newfile = new CPartFile(link);
	if (newfile->GetStatus() == PS_ERROR){
		delete newfile;
		return;
	}
	AddDownload(newfile);
    EMULE_CATCH
}

void CDownloadQueue::AddFileLinkToDownload(CED2KFileLink* pLink)
{
    EMULE_TRY
	CPartFile* newfile = new CPartFile(pLink);
	if (newfile->GetStatus() == PS_ERROR){
		delete newfile;
		newfile=NULL;
	}
	else AddDownload(newfile);
	if(pLink->HasValidSources()) {
		if (newfile) newfile->AddClientSources(pLink->SourcesList);
		else{
				CPartFile* partfile = GetFileByID((uchar*)pLink->GetHashKey());
				if (partfile) partfile->AddClientSources(pLink->SourcesList);
		}

	}
    EMULE_CATCH
}

//in the official they renamed this funcion StartNextFile()
// Added by Tarod
// Starts a paused download (takes in account priority)
void CDownloadQueue::StartPausedDownload() {
    EMULE_TRY
	CPartFile* cur_file ;
	CPartFile* file_to_start = NULL ;
	
	for (POSITION pos = filelist.GetHeadPosition() ; pos != (POSITION)NULL ; filelist.GetNext(pos)) {
		cur_file = (CPartFile*)filelist.GetAt(pos) ;
		if (cur_file->GetStatus() == PS_PAUSED) {
			if (file_to_start ==NULL || file_to_start->GetPriority() < cur_file->GetPriority()) {
				file_to_start = cur_file ;
			}
		}
	}

	if (file_to_start != NULL) file_to_start->ResumeFile() ;
    EMULE_CATCH
}

// Deactivates all downloading files A4AF auto setting
void CDownloadQueue::DisableAllA4AFAuto(void)
{
    EMULE_TRY
	for (POSITION pos = filelist.GetHeadPosition(); pos != NULL; filelist.GetNext(pos)) {
		CPartFile* cur_file = (CPartFile*)filelist.GetAt(pos);
		if (cur_file != NULL) cur_file->SetA4AFAuto(false);
	}
    EMULE_CATCH
}

// Returns A4AF auto file
//
// [Return]
//      Pointer fo file that is A4AF or NULL in case that there is no A4AF auto file.
CPartFile* CDownloadQueue::GetA4AFAutoFile(void)
{
    EMULE_TRY
	CPartFile* A4AF_auto_file = NULL;
	POSITION pos = filelist.GetHeadPosition();
	while (A4AF_auto_file == NULL && pos != NULL) {
		if (((CPartFile*)filelist.GetAt(pos))->IsA4AFAuto()) {
			A4AF_auto_file = (CPartFile*)filelist.GetAt(pos);
		}
		filelist.GetNext(pos);
	}

	return A4AF_auto_file;
    EMULE_CATCH
    return NULL;
}

void CDownloadQueue::AddDownload(CPartFile* newfile)
{
    EMULE_TRY
	filelist.AddTail(newfile);
	SortByPriority();
	// Added by Tarod
	// Checks if must add file paused and if not set first load saved sources time interval to 2 minutes
	// so we can get Exchanged sources excedings faster
	if (theApp.glob_prefs->StartDownloadPaused()) {
		newfile->StopFile() ;
	} else {
#ifdef OLD_SLS_ENABLED
		newfile->SetLoadSourcesTimeInterval(2 * 60 * 1000) ;
#endif //OLD_SLS_ENABLED
	}
	// End by Tarod
	theApp.emuledlg->transferwnd.downloadlistctrl.AddFile(newfile);
	theApp.emuledlg->AddLogLine(true,GetResString(IDS_NEWDOWNLOAD),newfile->GetFileName());
	CString msgTemp;
	msgTemp.Format(GetResString(IDS_NEWDOWNLOAD)+"\n",newfile->GetFileName());
	theApp.emuledlg->ShowNotifier(msgTemp, TBN_DLOAD_ADD);
    EMULE_CATCH
}

bool CDownloadQueue::IsFileExisting(uchar* fileid)
{
    EMULE_TRY
	if (CKnownFile* file = sharedfilelist->GetFileByID((uchar*)fileid)){
		if (file->IsPartFile())
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_ALREADY_DOWNLOADING), file->GetFileName());
		else
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_ALREADY_DOWNLOADED), file->GetFileName());
		return true;
	}
	else if ( file = GetFileByID((uchar*)fileid)){
		theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_ALREADY_DOWNLOADING), file->GetFileName());
		return true;
	}
    EMULE_CATCH
	return false;
}


void CDownloadQueue::Process()
{
    EMULE_TRY
	uint32 downspeed = 0;
	if (app_prefs->GetMaxDownload() != UNLIMITED && datarate > 1500){
		downspeed = (app_prefs->GetMaxDownload()*1024*100)/(datarate+1); //(uint16)((float)((float)(app_prefs->GetMaxDownload()*1024)/(datarate+1)) * 100);
		if (downspeed < 50)
			downspeed = 50;
		else if (downspeed > 200)
			downspeed = 200;
		//if (app_prefs->GetMaxDownload()*1024 < datarate)
		//	downspeed = 0xFFF;
	}
	datarate = 0;
	for (POSITION pos =filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file =  filelist.GetAt(pos);
		if ((cur_file->GetStatus() == PS_READY || cur_file->GetStatus() == PS_EMPTY) &&(cur_file->GetPriority() == PR_HIGH || cur_file->GetPriority() == PR_AUTO))	//jicxicmic fix
		{
			datarate += cur_file->Process(downspeed);
		}
	}
	for (POSITION pos =filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file =  filelist.GetAt(pos);
		if ((cur_file->GetStatus() == PS_READY || cur_file->GetStatus() == PS_EMPTY) && cur_file->GetPriority() == PR_NORMAL){
			datarate += cur_file->Process(downspeed);
		}
	}
	for (POSITION pos =filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file =  filelist.GetAt(pos);
		if ((cur_file->GetStatus() == PS_READY || cur_file->GetStatus() == PS_EMPTY) && cur_file->GetPriority() == PR_LOW){
			datarate += cur_file->Process(downspeed); 
		}
	}
	udcounter++;
	DWORD dwCurrentTime	=	GetTickCount(); // Use one variable to save few CPU cycles and have better time data [Lord KiRon]
	if (udcounter == 5){
		if((!lastudpstattime) || (dwCurrentTime - lastudpstattime) > UDPSERVERSTATTIME){
			lastudpstattime = dwCurrentTime;
			theApp.serverlist->ServerStats();
		}
	}
	if (udcounter == 10){
		udcounter = 0;
		if ((!lastudpsearchtime) || (dwCurrentTime - lastudpsearchtime) > UDPSERVERREASKTIME)
			SendNextUDPPacket();
	}


	//--- xrmb:keepPartFileStats ---
	//--- i hope that is a safe point to save the stats ---
	if(m_lastPartFileStatsSave+5*60*1000 < dwCurrentTime )
	{
		m_lastPartFileStatsSave=dwCurrentTime;

		SaveAllPartFileStats();
	}
	//--- :xrmb ---
    EMULE_CATCH
}

void	CDownloadQueue::SaveAllPartFileStats()
{
    EMULE_TRY
	for (POSITION pos =filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos))
	{
		CPartFile* cur_file =  filelist.GetAt(pos);
		cur_file->SavePartFileStats();
	}
    EMULE_CATCH
}

CPartFile*	CDownloadQueue::GetFileByID(uchar* filehash)
{
    EMULE_TRY
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		if (!memcmp(filehash,filelist.GetAt(pos)->GetFileHash(),16))
			return filelist.GetAt(pos);
	}
    EMULE_CATCH
	return 0;
}

bool CDownloadQueue::IsPartFile(void* totest)
{
    EMULE_TRY
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos))
		if (totest == filelist.GetAt(pos))
			return true;
    EMULE_CATCH
	return false;
}

void CDownloadQueue::CheckAndAddSource(CPartFile* sender,CUpDownClient* source, bool saveSources)
{
    EMULE_TRY
	// if we block loopbacks at this point it should prevent us from connecting to ourself
	if(!source->HasLowID() && (source->GetUserID() & 0xFF) == 0x7F) {
		delete source;
		return;
	}	
	// double check... just testing...
	// [Lord KiRon] v2
	if (memcmp(theApp.glob_prefs->GetUserHash(), source->GetUserHash(), 16) == 0)
	{
		delete source;
		return;
	} 
	// [/Lord KiRon] v2 

	// uses this only for temp. clients
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (int sl=0;sl<SOURCESSLOTS;sl++) if (!cur_file->srclists[sl].IsEmpty())
		for (POSITION pos2 = cur_file->srclists[sl].GetHeadPosition();pos2 != 0; cur_file->srclists[sl].GetNext(pos2)){
			if (cur_file->srclists[sl].GetAt(pos2)->Compare(source)){
				if (cur_file == sender){ // this file has already this source
					delete source;
					return;
				}
				// set request for this source
				if (cur_file->srclists[sl].GetAt(pos2)->AddRequestForAnotherFile(sender)){
					// add it to uploadlistctrl
					theApp.emuledlg->transferwnd.downloadlistctrl.AddSource(sender,cur_file->srclists[sl].GetAt(pos2),true);
					delete source;
					return;
				}
				else{
					delete source;
					return;
				}
			}
		}
	}
	//our new source is real new but maybe it is already uploading to us?
	//if yes the known client will be attached to the var "source"
	//and the old sourceclient will be deleted
	if (theApp.clientlist->AttachToAlreadyKnown(&source,0))
		source->reqfile = sender;
	else
		theApp.clientlist->AddClient(source,true);
	
	sender->srclists[source->sourcesslot].AddTail(source);
	theApp.emuledlg->transferwnd.downloadlistctrl.AddSource(sender,source,false);
	//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	UpdateDisplayedInfo();

	 //////////////////////////////////////////////////
     //code for saving the sources to disk

     if(saveSources && m_bKeepStoredSourcesOpened){
         //store the new source immediately to the CStoredSources-object of the part-file
#ifdef OLD_SLS_ENABLED
         sender->SaveToStoredSources(source);
#endif //OLD_SLS_ENABLED
     }//end of modification by Mr.Fry
     //////////////////////////////////////////////////
   EMULE_CATCH
}

BOOL CDownloadQueue::CheckAndAddKnownSource(CPartFile* sender,CUpDownClient* source)
{
    EMULE_TRY
	// SyruS "stop adding sources on stopped, completing and completed files" //Cax2: modified for current codebase....
	if(sender->GetStatus()==PS_PAUSED || sender->GetStatus() > 7)		//Cax2 + SyruS check
		return FALSE;
	// SyruS end
	if(!source->HasLowID() && (source->GetUserID() & 0xFF) == 0x7F)
		return FALSE;
	// [Lord KiRon] v3 
   	if (memcmp(theApp.glob_prefs->GetUserHash(), source->GetUserHash(), 16) == 0) 
   	{ 
    	// This is most important part of the fix 
        // if we not throw here, CClientReqSocket::ProcessPacket will never diconnect a client 
        /*throw CString(_T("Got myself as a source"));*/ 
        return FALSE; 
	}
    // [/Lord KiRon] v3 
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (int sl=0;sl<SOURCESSLOTS;sl++) if (!cur_file->srclists[sl].IsEmpty())
		if (cur_file->srclists[sl].Find(source)){
			if (cur_file == sender)
			{
				// [Lord KiRon] v3
				return TRUE;
				// [/Lord KiRon] v3 
			}
			if (source->AddRequestForAnotherFile(sender))
				theApp.emuledlg->transferwnd.downloadlistctrl.AddSource(sender,source,true);
			return TRUE;
		}
	}
	source->reqfile = sender;
	sender->srclists[source->sourcesslot].AddTail(source);
	theApp.emuledlg->transferwnd.downloadlistctrl.AddSource(sender,source,false);
	UpdateDisplayedInfo();
	//////////////////////////////////////////////////
	//code for saving the sources to disk

	if(m_bKeepStoredSourcesOpened){
		//store the new source immediately to the CStoredSources-object of the part-file
#ifdef OLD_SLS_ENABLED
		sender->SaveToStoredSources(source);
#endif //OLD_SLS_ENABLED
	}//end of modification by Mr.Fry
	//////////////////////////////////////////////////

	return TRUE; 
    EMULE_CATCH
    return FALSE;
}

bool CDownloadQueue::RemoveSource(CUpDownClient* toremove, bool	updatewindow)
{
    EMULE_TRY
	bool removed = false;
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (int sl=0;sl<SOURCESSLOTS;sl++) if (!cur_file->srclists[sl].IsEmpty())
		for (POSITION pos2 = cur_file->srclists[sl].GetHeadPosition();pos2 != 0; cur_file->srclists[sl].GetNext(pos2)){
			if (toremove == cur_file->srclists[sl].GetAt(pos2)){
				cur_file->srclists[sl].RemoveAt(pos2);
				cur_file->NewSrcPartsInfo();
				removed = true;
				break;
			}
		}
		cur_file->UpdateAvailablePartsCount();
	}
	if (updatewindow){
		toremove->SetDownloadState(DS_NONE);
		theApp.emuledlg->transferwnd.downloadlistctrl.RemoveSource(toremove,0);
	}
	toremove->reqfile = 0;
	return removed;
    EMULE_CATCH
    return false;
}

void CDownloadQueue::RemoveFile(CPartFile* toremove)
{
    EMULE_TRY
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		if (toremove == filelist.GetAt(pos)){
			filelist.RemoveAt(pos);
			return;
		}
	}
    EMULE_CATCH
}

void CDownloadQueue::DeleteAll()
{
    EMULE_TRY
	POSITION pos;
	for (pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (int sl=0;sl<SOURCESSLOTS;sl++) if (!cur_file->srclists[sl].IsEmpty())
			cur_file->srclists[sl].RemoveAll();
		// Barry - Should also remove all requested blocks
		// Don't worry about deleting the blocks, that gets handled 
		// when CUpDownClient is deleted in CClientList::DeleteAll()
		cur_file->RemoveAllRequestedBlocks();
	}
    EMULE_CATCH
}

bool CDownloadQueue::SendNextUDPPacket()
{
	EMULE_TRY
	if (filelist.IsEmpty() || !theApp.serverconnect->IsConnected())
		return false;
	if (!cur_udpserver)
		if (!(cur_udpserver = theApp.serverlist->GetNextServer(cur_udpserver))){
			TRACE("ERROR:SendNextUDPPacket() no server found");
			StopUDPRequests();
		};

	// get nextfile
	CPartFile* nextfile = 0;
	while (!(nextfile && (nextfile->GetStatus() == PS_READY ||nextfile->GetStatus() == PS_EMPTY))){
		if (lastfile == 0){
			nextfile = filelist.GetHead();
			lastfile = nextfile;
		}
		else{
			POSITION pos = filelist.Find(lastfile);
			if (!pos){
				TRACE("Error: CDownloadQueue::SendNextUDPPacket()");
				nextfile = filelist.GetHead();
				lastfile = nextfile;
			}
			else{
				filelist.GetNext(pos);
				if (pos == 0){
					cur_udpserver = theApp.serverlist->GetNextServer(cur_udpserver);
					if (cur_udpserver == 0){
						//TRACE("finished");
						lastudpsearchtime = ::GetTickCount();
						lastfile = 0;
						return false; // finished (processed all file & all servers)
					}
					nextfile = filelist.GetHead();
					lastfile = nextfile;
				}
				else{
					nextfile = filelist.GetAt(pos);
					lastfile = nextfile;
				}
			}
		}
	}
	if( (theApp.glob_prefs->GetMaxSourcePerFileUDP()) < nextfile->GetSourceCount()) //<<--
		return true; 
	Packet packet(OP_GLOBGETSOURCES,16);
	memcpy(packet.pBuffer,nextfile->GetFileHash(),16);
	if (cur_udpserver != theApp.serverlist->GetServerByAddress(theApp.serverconnect->GetCurrentServer()->GetAddress(),theApp.serverconnect->GetCurrentServer()->GetPort()))
		theApp.serverconnect->SendUDPPacket(&packet,cur_udpserver,false);
	return true;
	EMULE_CATCH
	return false;
}

void CDownloadQueue::StopUDPRequests()
{
	EMULE_TRY
	cur_udpserver = 0;
	lastudpsearchtime = ::GetTickCount();
	lastfile = 0;
	EMULE_CATCH
}

void CDownloadQueue::SortByPriority()
{
	EMULE_TRY
	POSITION pos1, pos2;
	uint16 i = 0;
	for( pos1 = filelist.GetHeadPosition(); ( pos2 = pos1 ) != NULL; ){
	   filelist.GetNext(pos1);
	   CPartFile* cur_file = filelist.GetAt(pos2);
	   if (cur_file->GetPriority() == PR_HIGH){
		   filelist.AddHead(cur_file);
		   filelist.RemoveAt(pos2);
	   }
	   else if (cur_file->GetPriority() == PR_LOW){
		   filelist.AddTail(cur_file);
		   filelist.RemoveAt(pos2);
	   }
	   i++;
	   if (i == filelist.GetCount())
		   break;
   }
	EMULE_CATCH
}

void CDownloadQueue::GetDownloadStats(int results[]) 
{
	EMULE_TRY
	
	results[0]=0;
	results[1]=0;

	for (POSITION pos =theApp.downloadqueue->filelist.GetHeadPosition();pos != 0;theApp.downloadqueue->filelist.GetNext(pos)){
		CPartFile* cur_file =  filelist.GetAt(pos);
		results[0]+=cur_file->GetSourceCount();
		results[1]+=cur_file->GetTransferingSrcCount();
	}
	EMULE_CATCH
}

CUpDownClient* CDownloadQueue::GetDownloadClientByIP(uint32 dwIP)
{
	EMULE_TRY
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* cur_file = filelist.GetAt(pos);
		for (int sl=0;sl<SOURCESSLOTS;sl++) if (!cur_file->srclists[sl].IsEmpty())
		for (POSITION pos2 = cur_file->srclists[sl].GetHeadPosition();pos2 != 0; cur_file->srclists[sl].GetNext(pos2)){
			if (dwIP == cur_file->srclists[sl].GetAt(pos2)->GetIP()){
				return cur_file->srclists[sl].GetAt(pos2);
			}
		}
	}
	EMULE_CATCH
	return NULL;
}

void CDownloadQueue::GetDownloadStats(int results[],uint64& pui64TotFileSize,uint64& pui64TotBytesLeftToTransfer,uint64& pui64TotNeededSpace) 
{
	EMULE_TRY	
	results[0]=0;
	results[1]=0;
	results[2]=0;

	for (POSITION pos =theApp.downloadqueue->filelist.GetHeadPosition();pos != 0;theApp.downloadqueue->filelist.GetNext(pos)){
		uint32 ui32SizeToTransfer=0;
		uint32 ui32NeededSpace=0;
		CPartFile* cur_file =  filelist.GetAt(pos);
		if (cur_file->GetPartfileStatus() != GetResString(IDS_PAUSED)) { 
		cur_file->GetSizeToTransferAndNeededSpace (ui32SizeToTransfer,ui32NeededSpace);
        pui64TotFileSize += cur_file->GetFileSize(); 
        pui64TotBytesLeftToTransfer += ui32SizeToTransfer;
        pui64TotNeededSpace += ui32NeededSpace;
        results[2]++; 
		} 
		results[0]+=cur_file->GetSourceCount();
		results[1]+=cur_file->GetTransferingSrcCount();
	}
	EMULE_CATCH
}

// added by FoRcHa: ////////////////////////
int CDownloadQueue::GetFileCount()
{
	EMULE_TRY
	return filelist.GetCount();
	EMULE_CATCH
	return 0;
}

int CDownloadQueue::GetTransferingFiles()
{
	EMULE_TRY
	int iCount = 0;
	for(POSITION pos = theApp.downloadqueue->filelist.GetHeadPosition(); pos != 0; theApp.downloadqueue->filelist.GetNext(pos))
	{
		CPartFile* cur_file = filelist.GetAt(pos);
		if(cur_file->GetTransferingSrcCount() > 0)
			iCount++;
	}
	return iCount;
	EMULE_CATCH
	return 0;
}
/////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////
//Storing Sources:
#ifdef OLD_SLS_ENABLED
bool CDownloadQueue::SaveAllSourcesToFile(bool debug)
{
	EMULE_TRY
	bool errorOccured=false;
	for (POSITION pos =filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos)){
		CPartFile* partFile = filelist.GetAt(pos);
		if(!partFile->SaveSources()){
			TRACE("Could not save sources (clients) for one downloading file\n\t(probably there are no sources for this file)\n");			
			errorOccured = true;
		}
	}
	if(errorOccured) return false;
	return true;
	EMULE_CATCH
	return false;
}
#endif //OLD_SLS_ENABLED
bool CDownloadQueue::IsMaxClientsExceeded()
{
	EMULE_TRY
	return (theApp.clientlist->GetClientCount() > (int)theApp.glob_prefs->GetMaxConnections());
	EMULE_CATCH
	return false;
}
///////////