//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "partfile.h"
#include "emule.h"
#include "updownclient.h"
#include <math.h>
#include "ED2KLink.h"
#include "Preview.h"
#include "ini2.h"
#include "ArchiveRecovery.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// Barry - use this constant for both places
#define PROGRESS_HEIGHT 3

CBarShader CPartFile::s_LoadBar(PROGRESS_HEIGHT); // Barry - was 5
CBarShader CPartFile::s_ChunkBar(16); 
CMutex CPartFile::s_ComplitingMutex;

CPartFile::CPartFile()
{
	Init();
}

CPartFile::CPartFile(CSearchFile* searchresult)
{
	Init();

	EMULE_TRY

	memcpy(filehash,searchresult->GetFileHash(),16);
	for (int i = 0; i != searchresult->taglist.GetCount();i++){
		switch (searchresult->taglist[i]->tag->specialtag){
			case FT_FILENAME:{
				filename = nstrdup(searchresult->taglist[i]->tag->stringvalue);
				break;
			}
			case FT_FILESIZE:{
				filesize = searchresult->taglist[i]->tag->intvalue;
				break;
			}
			default:
				CTag* newtag = new CTag(searchresult->taglist[i]->tag);
				taglist.Add(newtag);
		}
	}
	CreatePartFile();

	EMULE_CATCH
}

CPartFile::CPartFile(CString edonkeylink)
{
	EMULE_TRY

	CED2KLink* pLink = 0;
	try {
		pLink = CED2KLink::CreateLinkFromUrl(edonkeylink);
		_ASSERT( pLink != 0 );
		CED2KFileLink* pFileLink = pLink->GetFileLink();
		if (pFileLink==0) 
			throw GetResString(IDS_ERR_NOTAFILELINK);
		InitializeFromLink(pFileLink);
	} catch (CString error) {
		OUTPUT_DEBUG_TRACE();
		char buffer[200];
		sprintf(buffer,GetResString(IDS_ERR_INVALIDLINK),error.GetBuffer());
		theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_LINKERROR), buffer);
		status = PS_ERROR;
	}
	delete pLink;

	EMULE_CATCH
}

void CPartFile::InitializeFromLink(CED2KFileLink* fileLink)
{
	Init();

	EMULE_TRY

	try{
		filename = nstrdup( fileLink->GetName() );
		filesize = fileLink->GetSize();
		memcpy(filehash,fileLink->GetHashKey(),sizeof(filehash));
		if (!theApp.downloadqueue->IsFileExisting(filehash))
			CreatePartFile();
		else
			status = PS_ERROR;
		}
	catch(CString error){
		OUTPUT_DEBUG_TRACE();
		char buffer[200];
		sprintf(buffer,GetResString(IDS_ERR_INVALIDLINK),error.GetBuffer());
		theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_LINKERROR), buffer);
		status = PS_ERROR;
	}

	EMULE_CATCH
}

CPartFile::CPartFile(CED2KFileLink* fileLink)
{
	InitializeFromLink(fileLink);
}

void CPartFile::Init()
{
	EMULE_TRY

	fullname = 0;
	newdate = true;
	lastsearchtime = 0;
	lastpurgetime = GetTickCount();
	paused = false;
	status = PS_EMPTY;
	transfered = 0;
	priority = PR_NORMAL;
	srcarevisible = false;
	transferingsrc = 0;
	datarate = 0;
	hashsetneeded = true;
	count = 0;
	percentcompleted = 0;
	partmetfilename = 0;
	completedsize=0;

	/////////////////////////////////////
	//for saving sources:
	
#ifdef OLD_SLS_ENABLED
	m_bLoadedSourcesCompletely				= true;//do load sources only on a call of LoadAndAddSourceFromFile(...)
	m_iSourceIndex							= 0;
	m_iSourceIDIndex						= -1;
	m_bWasConnectedToServerLastTime			= false;
	m_iLastTimeSourcesLoaded				= ::GetTickCount();	//I'm not sure, maybe it should be 0
	m_iLastTimeSourcesSaved					= ::GetTickCount();	//I'm not sure, maybe it should be 0
	m_iLastTimeSourcesLoadedPartial			= ::GetTickCount();	//I'm not sure, maybe it should be 0
	m_iLoadSourcesAtOnceLimit				= RESTORE_SOURCES_AT_ONCE;
	m_iLoadingSourcesTimeInterval			= RESTORE_SOURCES_TIME;
	m_iLoadingSourcesPartialTimeInterval	= RESTORE_SOURCES_SLOWLY_TIME;
	m_iSavingSourcesTimeInterval			= STORE_SOURCES_TIME;
	m_stscrStoringSources					= NULL;
	m_bLoadSavedSources						= true;
#endif //OLD_SLS_ENABLED
	m_bPreviewing=false;
	lastseencomplete=NULL;
	availablePartsCount=0;
	m_ClientSrcAnswered = 0;
	m_LastNoNeededCheck = 0;
	m_iRate=0;
	m_strComment="";
	m_nTotalBufferData = 0;
	m_nLastBufferFlushTime = 0;
	m_bPercentUpdated = false;
	m_bRecoveringArchive = false;
	m_iGainDueToCompression = 0;
	m_iLostDueToCorruption = 0;
	m_iTotalPacketsSavedDueToICH = 0;
	m_nSavedReduceDownload = 0;
	m_is_A4AF_auto = false;
	hasRating=false;
	hasComment=false;

	// Tarod
	if (theApp.glob_prefs->IsDAPEnabled())
		m_bAutoPriority = true;
	else
		m_bAutoPriority = false;
	m_LastRemovedTime = 0;

    m_lastDownTransfer = CTime::GetCurrentTime();

	EMULE_CATCH
}

CPartFile::~CPartFile()
{
	EMULE_TRY

	// Barry - Ensure all buffered data is written
	FlushBuffer();

	if (fullname)
		delete[] fullname;
	if (partmetfilename)
		delete[] partmetfilename;
	if(tempdir) // FoRcHa
		delete[] tempdir;
	m_SrcpartFrequency.RemoveAll();
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos))
		delete gaplist.GetAt(pos);
	/////////////////////////////////////
	//for saving sources: mr fry
#ifdef OLD_SLS_ENABLED
	if(m_stscrStoringSources != NULL) {
		delete m_stscrStoringSources;
		m_stscrStoringSources = NULL;
	}
#endif //OLD_SLS_ENABLED
	EMULE_CATCH
}

void CPartFile::CreatePartFile()
{
	EMULE_TRY

	tempdir = nstrdup(theApp.glob_prefs->GetTempDir());	// added by InterCeptor (multiple tempdirs) 19.11.02
	if (tempdir[strlen(tempdir)-1] == '\\') tempdir[strlen(tempdir)-1] = 0;	// added by InterCeptor (check for \\) 19.11.02
  	int i = 0; 
	CString filename; 
	do 
	{ 
		i++; 
		filename.Format("%s\\%03i.part", tempdir, i); 
	
	}while(PathFileExists(filename)); 
	partmetfilename = new char[15]; 
	sprintf(partmetfilename,"%03i.part.met",i); 

	fullname = new char[strlen(tempdir)+strlen(partmetfilename)+MAX_PATH];
	sprintf(fullname,"%s\\%s",tempdir,partmetfilename);
	char* buffer = nstrdup(partmetfilename);
	buffer[strlen(buffer)-4] = 0;
	CTag* partnametag = new CTag(FT_PARTFILENAME,buffer);
	delete[] buffer;
	taglist.Add(partnametag);
	
	Gap_Struct* gap = new Gap_Struct;
	gap->start = 0;
	gap->end = filesize-1;
	gaplist.AddTail(gap);

	char* partfull = nstrdup(fullname);
	partfull[strlen(partfull)-4] = 0;
	
	
	if (!m_hpartfile.Open(partfull,CFile::modeCreate|CFile::modeReadWrite|CFile::shareDenyWrite|CFile::osSequentialScan)){
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_CREATEPARTFILE));
		status = PS_ERROR;
	}
	delete[] partfull;
	m_SrcpartFrequency.SetSize(GetPartCount());
	for (uint32 i = 0; i != GetPartCount();i++)
		m_SrcpartFrequency.Add(0);
	paused = false;

	// Before we save file we need to stablish UAP by Tarod
	if (theApp.glob_prefs->IsUAPEnabled()) {
		CKnownFile::SetAutoPriority(true);
		CKnownFile::UpdateUploadAutoPriority();
		theApp.emuledlg->sharedfileswnd.sharedfilesctrl.UpdateItem((CKnownFile*)this);
	}
	// End Tarod
	
	SavePartFile();

	EMULE_CATCH
}

bool CPartFile::LoadPartFile(char* in_directory,char* in_filename)
{
	EMULE_TRY

	tempdir = nstrdup(in_directory);	// added by InterCeptor (multiple tempdirs) 19.11.02
    CMap<uint16,uint16,Gap_Struct*,Gap_Struct*> gap_map; // safe Load Gaps [SlugFiller]
	transfered = 0;
	partmetfilename = nstrdup(in_filename);
	directory = nstrdup(in_directory);
	char* buffer = new char[strlen(directory)+strlen(partmetfilename)+2];
	if (directory[strlen(directory)-1] == '\\') directory[strlen(directory)-1] = 0;	// added by InterCeptor (check for \\) 19.11.02
	sprintf(buffer,"%s\\%s",directory,partmetfilename);
	fullname = buffer;
	CSafeFile file;
	try{
		// readfile data form part.met file
		if (!file.Open(fullname,CFile::modeRead)){
			theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_OPENMET),partmetfilename,filename);
			return false;
		}
		uint8 version;
		file.Read(&version,1);
		if (version != PARTFILE_VERSION){
			if (file!= INVALID_HANDLE_VALUE) file.Close();
			theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_BADMETVERSION),partmetfilename,filename);
			return false;
		}
		LoadDateFromFile(&file);
		LoadHashsetFromFile(&file,false);
		LoadMovieMode(); // DonGato - Moview Preview Mode

		uint32 tagcount;
		file.Read(&tagcount,4);
		for (uint32 j = 0; j != tagcount;j++){
			CTag* newtag = new CTag(&file);
			switch(newtag->tag->specialtag){
				case FT_FILENAME:{
					if(newtag->tag->stringvalue == NULL) {
						theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_METCORRUPT),partmetfilename,filename);
						delete newtag;
						return false;
					}
					filename = nstrdup(newtag->tag->stringvalue);
					delete newtag;
					break;
				}
				case FT_LASTSEENCOMPLETE: {
					lastseencomplete = newtag->tag->intvalue;
					delete newtag;
					break;
				}
				case FT_FILESIZE:{
					filesize = newtag->tag->intvalue;
					delete newtag;
					break;
				}
				case FT_TRANSFERED:{
					transfered = newtag->tag->intvalue;
					delete newtag;
					break;
				}
				case FT_PRIORITY:{
					priority = newtag->tag->intvalue;
					if ((priority==PR_AUTO)&&(!theApp.glob_prefs->IsUAPEnabled())) priority=PR_NORMAL;
					delete newtag;

					if (priority == PR_AUTO)
					{ 
						SetAutoPriority(true);
						CKnownFile::UpdateUploadAutoPriority();
					}
					else
						SetAutoPriority(false);
					break;
				}
				case FT_STATUS:{
					paused = newtag->tag->intvalue;
					delete newtag;
					break;
				}
				case FT_ULPRIORITY: {
					if ((priority==PR_AUTO)&&(!theApp.glob_prefs->IsUAPEnabled())) priority=PR_NORMAL;
					if ((uint8)newtag->tag->intvalue == PR_AUTO) {
						CKnownFile::SetAutoPriority(true);
						CKnownFile::UpdateUploadAutoPriority();
					} else {
						CKnownFile::SetAutoPriority(false);
						CKnownFile::SetPriority((uint8)newtag->tag->intvalue);
					}
					delete newtag;
					break;
				}
				default:{
					// Start Changes by Slugfiller for better exception handling
					if ((!newtag->tag->specialtag) &&
						 (newtag->tag->tagname[0] == FT_GAPSTART ||
					     newtag->tag->tagname[0] == FT_GAPEND)){
						Gap_Struct* gap;
						uint16 gapkey = atoi(&newtag->tag->tagname[1]);
						if (!gap_map.Lookup(gapkey, gap))
						{
							gap = new Gap_Struct;
							gap_map.SetAt(gapkey, gap);
							gap->start = -1;
							gap->end = -1;
						}
						if (newtag->tag->tagname[0] == FT_GAPSTART)
							gap->start = newtag->tag->intvalue;
						if (newtag->tag->tagname[0] == FT_GAPEND)
							gap->end = newtag->tag->intvalue-1;
						delete newtag;
					// End Changes by Slugfiller for better exception handling
					}
					else
						taglist.Add(newtag);
				}
					
			}

		}
		if (file != INVALID_HANDLE_VALUE) file.Close();
	}
	catch(CFileException* error){
		OUTPUT_DEBUG_TRACE();
		if (error->m_cause == CFileException::endOfFile)
			theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_METCORRUPT),partmetfilename,filename);
		else{
			char buffer[150];
			error->GetErrorMessage(buffer,150);
			theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_FILEERROR),partmetfilename,filename,error);
		}
		error->Delete();
		return false;
	}

	// Now to flush the map into the list (Slugfiller)
	for (POSITION pos = gap_map.GetStartPosition(); pos != NULL; ){
		Gap_Struct* gap;
		uint16 gapkey;
 		gap_map.GetNextAssoc(pos, gapkey, gap);
		if (gap->start >= 0 && gap->end >=0 && gap->start <= gap->end)
        		gaplist.AddTail(gap); // All tags accounted for
    		else
     			delete gap; // Some of the tags were missing
    	}

	EMULE_TRY
	//--- xrmb:keepPartFileStats ---
	//--- i hope that is a safe point to load the stats ---
	LoadPartFileStats();
	//--- :xrmb ---
	EMULE_CATCH

	//check if this is a backup
	if(stricmp(strrchr(fullname, '.'), ".backup") == 0) {
		char *shorten = strrchr(fullname, '.');
		*shorten = 0;
		fullname = (char*)realloc(fullname, strlen(fullname) + 1);
	}

	// open permanent handle
	char* searchpath = nstrdup(fullname);
	searchpath[strlen(fullname)-4] = 0;
	if (!m_hpartfile.Open(searchpath,CFile::modeReadWrite|CFile::shareDenyWrite|CFile::osSequentialScan)){
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_FILEOPEN),fullname,filename);
		delete[] searchpath;
		return false;
	}
	delete[] searchpath;
			

	m_SrcpartFrequency.SetSize(GetPartCount());
	for (uint32 i = 0; i != GetPartCount();i++)
		m_SrcpartFrequency.Add(0);
	status = PS_READY;
	// check hashcount, filesatus etc
	if (hashlist.GetCount() < GetPartCount()){
		status = PS_EMPTY;
		hashsetneeded = true;
		return true;
	}
	else {
		hashsetneeded = false;
		bool empty = true;
		for (int i = 0; i != hashlist.GetSize(); i++){
			if (IsComplete(i*PARTSIZE,((i+1)*PARTSIZE)-1))
				empty = false;
		}
		if (empty)
			status = PS_EMPTY;
	}

	if (gaplist.IsEmpty()){	// is this file complete already?
		CompleteFile(false);
		return true;
	}

	// check date of .part file - if its wrong, rehash file
		CFileStatus filestatus;
	m_hpartfile.GetStatus(filestatus); // this; "...returns m_attribute without high-order flags" indicates a known MFC bug, wonder how many unknown there are... :)
        m_lastDownTransfer = filestatus.m_mtime;
		if (date != mktime(filestatus.m_mtime.GetLocalTm())){
			theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_REHASH),buffer,filename);
			// rehash
			status = PS_WAITINGFORHASH;
		    CAddFileThread* addfilethread = (CAddFileThread*) AfxBeginThread(RUNTIME_CLASS(CAddFileThread), THREAD_PRIORITY_NORMAL,0, CREATE_SUSPENDED);
		    addfilethread->SetValues(0,directory,m_hpartfile.GetFileName().GetBuffer(),this);
		    addfilethread->ResumeThread();
		}

	UpdateCompletedInfos();
	if ( completedsize > transfered )
		m_iGainDueToCompression = completedsize - transfered;
	else if ( completedsize != transfered )
		m_iLostDueToCorruption = transfered - completedsize;

	return true;

	EMULE_CATCH

	return false;
}

bool CPartFile::SavePartFile()
{
	EMULE_TRY

	switch (status){
		case PS_WAITINGFORHASH:
		case PS_HASHING:
			return false;
	}
	FILE* file = 0;
	try{
		//backup part.met file
		CString bakname;
		bakname.Format("%s\\%s.bak", theApp.glob_prefs->GetTempDir(), this->GetPartMetFileName());
		::CopyFile(fullname, bakname, FALSE);
		
		//get filedate
		CFileFind ff;
		char* searchpath = nstrdup(fullname);
		searchpath[strlen(fullname)-4] = 0;
		bool end = !ff.FindFile(searchpath,0);
		delete[] searchpath;
		if (!end)
				ff.FindNextFile();
		if (end || ff.IsDirectory())
			throw GetResString(IDS_ERR_PART_FNF);
		CTime lwtime;
		ff.GetLastWriteTime(lwtime);
		date = mktime(lwtime.GetLocalTm());
		ff.Close();
		uint32 lsc = mktime(lastseencomplete.GetLocalTm());
		CString backupName(fullname);
		backupName.Append(".backup");

		remove(backupName);
		rename(fullname, backupName);

		// readfile data form part.met file
		file = fopen(fullname,"wbS");
		if (!file)
			throw GetResString(IDS_ERR_OPENMETFILE);
		//version
		uint8 version = PARTFILE_VERSION;
		fwrite(&version,1,1,file);
		//date
		fwrite(&date,4,1,file);
		//hash
		fwrite(&filehash,16,1,file);
		uint16 parts = hashlist.GetCount();
		fwrite(&parts,2,1,file);
		for (int x = 0; x != parts; x++)
			fwrite(hashlist[x],16,1,file);
		//tags
		uint32 tagcount = taglist.GetCount()+7+(gaplist.GetCount()*2);
		fwrite(&tagcount,4,1,file);
		CTag* nametag = new CTag(FT_FILENAME,filename);
		nametag->WriteTagToFile(file);
		delete nametag;
		CTag* sizetag = new CTag(FT_FILESIZE,filesize);
		sizetag->WriteTagToFile(file);
		delete sizetag;
		CTag* transtag = new CTag(FT_TRANSFERED,transfered);
		transtag->WriteTagToFile(file);
		delete transtag;
		CTag* statustag = new CTag(FT_STATUS,(paused)? 1:0);
		statustag->WriteTagToFile(file);
		delete statustag;
		//UAP Hunter
		CTag* prioritytag ;
		if (IsAutoPrioritized())
			prioritytag = new CTag(FT_PRIORITY,(uint8)PR_AUTO);
		else 
			prioritytag = new CTag(FT_PRIORITY,priority);
		prioritytag->WriteTagToFile(file);
		delete prioritytag;
		//end UAP
		CTag* lsctag = new CTag(FT_LASTSEENCOMPLETE,lsc);
		lsctag->WriteTagToFile(file);
		delete lsctag;

		// Modified by Tarod for UAP
		CTag* ulprioritytag;
		if (CKnownFile::IsAutoPrioritized())
			ulprioritytag = new CTag(FT_ULPRIORITY, PR_AUTO);
		else
			ulprioritytag = new CTag(FT_ULPRIORITY,m_iPriority);
		ulprioritytag->WriteTagToFile(file);
		delete ulprioritytag;
		// End by Tarod
		for (uint32 j = 0; j != taglist.GetCount();j++)
			taglist[j]->WriteTagToFile(file);
		//gaps
		char* namebuffer = new char[10];
		char* number = &namebuffer[1];
		uint16 i_pos = 0;
		for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos)){
			itoa(i_pos,number,10);
			namebuffer[0] = FT_GAPSTART;
			CTag* gapstarttag = new CTag(namebuffer,gaplist.GetAt(pos)->start);
			gapstarttag->WriteTagToFile(file);
			// gap start = first missing byte but gap ends = first non-missing byte in edonkey
			// but I think its easier to user the real limits
			namebuffer[0] = FT_GAPEND;
			CTag* gapendtag = new CTag(namebuffer,(gaplist.GetAt(pos)->end)+1);
			gapendtag->WriteTagToFile(file);
			delete gapstarttag;
			delete gapendtag;
			i_pos++;
		}
		delete[] namebuffer;

		if (ferror(file))
			throw CString("unexpected write error");

		remove(backupName);
	}	
	catch(char* error){
		OUTPUT_DEBUG_TRACE();
		if (file)
			fclose(file);
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_SAVEMET),error,partmetfilename,filename);
		return false;
	}
	catch(CString error){
		OUTPUT_DEBUG_TRACE();
		if (file)
	fclose(file);
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_SAVEMET),error.GetBuffer(),partmetfilename,filename);
		return false;
	}
	fclose(file);
	return true;

	EMULE_CATCH

	return false;
}

void CPartFile::PartFileHashFinished(CKnownFile* result)
{
	EMULE_TRY

	newdate = true;
	bool errorfound = false;
	for (uint32 i = 0; i != hashlist.GetSize(); i++){
		if (IsComplete(i*PARTSIZE,((i+1)*PARTSIZE)-1)){
			if (!(result->GetPartHash(i) && !memcmp(result->GetPartHash(i),GetPartHash(i),16))){
				theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_FOUNDCORRUPTION),i+1,filename);		
				AddGap(i*PARTSIZE,((((i+1)*PARTSIZE)-1) >= filesize) ? filesize-1 : ((i+1)*PARTSIZE)-1);
				errorfound = true;
			}
		}
	}
	delete result;
	if (!errorfound){
		if (status == PS_COMPLETING){
			CompleteFile(true);
			return;
		}
		else
			theApp.emuledlg->AddLogLine(false,GetResString(IDS_HASHINGDONE),filename);
	}
	else{
		status = PS_READY;
		SavePartFile();
		return;
	}
	status = PS_READY;
	SavePartFile();
	theApp.sharedfiles->SafeAddKFile(this);

	EMULE_CATCH
}

void CPartFile::AddGap(uint32 start, uint32 end)
{
	EMULE_TRY

	POSITION pos1, pos2;
	for (pos1 = gaplist.GetHeadPosition();(pos2 = pos1) != NULL;){
		Gap_Struct* cur_gap = gaplist.GetNext(pos1);
		if (cur_gap->start >= start && cur_gap->end <= end){ // this gap is inside the new gap - delete
			gaplist.RemoveAt(pos2);
			delete cur_gap;
			continue;
		}
		else if (cur_gap->start >= start && cur_gap->start <= end){// a part of this gap is in the new gap - extend limit and delete
			end = cur_gap->end;
			gaplist.RemoveAt(pos2);
			delete cur_gap;
			continue;
		}
		else if (cur_gap->end <= end && cur_gap->end >= start){// a part of this gap is in the new gap - extend limit and delete
			start = cur_gap->start;
			gaplist.RemoveAt(pos2);
			delete cur_gap;
			continue;
		}
		else if (start >= cur_gap->start && end <= cur_gap->end){// new gap is already inside this gap - return
			return;
		}
	}
	Gap_Struct* new_gap = new Gap_Struct;
	new_gap->start = start;
	new_gap->end = end;
	gaplist.AddTail(new_gap);
	//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	UpdateDisplayedInfo();
	newdate = true;

	EMULE_CATCH
}

bool CPartFile::IsComplete(uint32 start, uint32 end)
{
	EMULE_TRY

	if (end >= filesize)
		end = filesize-1;
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos)){
		Gap_Struct* cur_gap = gaplist.GetAt(pos);
		if ((cur_gap->start >= start && cur_gap->end <= end)||(cur_gap->start >= start 
			&& cur_gap->start <= end)||(cur_gap->end <= end && cur_gap->end >= start)
			||(start >= cur_gap->start && end <= cur_gap->end)){
				return false;	
			}
	}
	return true;

	EMULE_CATCH

	return false;
}

bool CPartFile::IsPureGap(uint32 start, uint32 end)
{
	EMULE_TRY

	if (end >= filesize)
		end = filesize-1;
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos)){
		Gap_Struct* cur_gap = gaplist.GetAt(pos);
		if (start >= cur_gap->start  && end <= cur_gap->end ){
			return true;
		}
	}

	EMULE_CATCH

	return false;
}

bool CPartFile::IsAlreadyRequested(uint32 start, uint32 end)
{
	EMULE_TRY

	for (POSITION pos =  requestedblocks_list.GetHeadPosition();pos != 0; requestedblocks_list.GetNext(pos)){
		Requested_Block_Struct* cur_block =  requestedblocks_list.GetAt(pos);
		if ((start <= cur_block->EndOffset) && (end >= cur_block->StartOffset))
			return true;
	}

	EMULE_CATCH

	return false;
}

/* rewritten
//Vorlost for LoadRarestFinishedChunks Patch2
bool CPartFile::GetNextEmptyBlockInPart(uint16 partnumber,Requested_Block_Struct* result)
{
	EMULE_TRY

	bool forward = true;
	uint32 lastblock = ceil((float)PARTSIZE/BLOCKSIZE);
	uint32 lastblockinpart = lastblock;
	uint32 start,end,i;
	char buffer[200] = "";

	//Choose FW or BW - Mode
	if (partnumber == GetPartCount()-2)
		end = (PARTSIZE*partnumber) + (lastblock*BLOCKSIZE)-1; //Endadress from 1chunk before the last chunk
	if (partnumber == GetPartCount()-1   //Is this the last chunk ?
		|| ((partnumber == GetPartCount()-2) //Or is this 1chunk before the last chunk ?
		&& (((GetFileSize()-1) - MINLASTMOVIESIZE) <= end )))  //Is 1/8 from fileend part of 1chunk before the last chunk ?
		forward = false;

	//Get last possible blocknumber
	// if ( !forward ){
	for (lastblockinpart = 0; lastblockinpart != lastblock;lastblockinpart++)
	{
		start = (PARTSIZE*partnumber) + lastblockinpart*BLOCKSIZE;
		if (start >= GetFileSize())
			break;
	}

	if ( forward )
	{
		for (i = 0; i != lastblock;i++)
		{
			start = (PARTSIZE*partnumber) + i*BLOCKSIZE;
			end  = (PARTSIZE*partnumber) + ((i+1)*BLOCKSIZE)-1;
			if (end >= (uint32)PARTSIZE*(partnumber+1))
				end = (PARTSIZE*(partnumber+1))-1;
			if (start >= GetFileSize())
				break;
			if (end >= GetFileSize())
				end = GetFileSize()-1;
			if ( (!IsComplete(start,end)) && !IsAlreadyRequested(start,end))
			{
				if (result)
				{
					result->StartOffset = start;
					result->EndOffset = end;
					memcpy(result->FileID,GetFileHash(),16);
				}
				return true;
			}
		}
		return false;
	}
	else
	{
		for (i = lastblockinpart-1; i != -1;i--)
		{
			start = (PARTSIZE*partnumber) + i*BLOCKSIZE;
			end  = (PARTSIZE*partnumber) + ((i+1)*BLOCKSIZE)-1;
			if (end >= (uint32)PARTSIZE*(partnumber+1))
				end = (PARTSIZE*(partnumber+1))-1;
			if (start >= GetFileSize())
				break;
			if (end >= GetFileSize())
				end = GetFileSize()-1;
			if ( (!IsComplete(start,end)) && !IsAlreadyRequested(start,end))
			{
				if (result)
				{
					result->StartOffset = start;
					result->EndOffset = end;
					memcpy(result->FileID,GetFileHash(),16);
				}
				return true;
			}
		}
		return false;
	}

	EMULE_CATCH

	return false;
}
// End Vorlost
*/

bool CPartFile::GetNextEmptyBlockInPart(uint16 partNumber, Requested_Block_Struct *result)
{
	EMULE_TRY

	Gap_Struct *firstGap;
	Gap_Struct *currentGap;
	uint32 end;
	uint32 blockLimit;

	// Find start of this part
	uint32 partStart = (PARTSIZE * partNumber);
	uint32 start = partStart;

	// What is the end limit of this block, i.e. can't go outside part (or filesize)
	uint32 partEnd = (PARTSIZE * (partNumber + 1)) - 1;
	if (partEnd >= GetFileSize())
		partEnd = GetFileSize() - 1;

	// Loop until find a suitable gap and return true, or no more gaps and return false
	while (true)
	{
		firstGap = NULL;

		// Find the first gap from the start position
		for (POSITION pos = gaplist.GetHeadPosition(); pos != 0; gaplist.GetNext(pos))
		{
			currentGap = gaplist.GetAt(pos);
			// Want gaps that overlap start<->partEnd
			if ((currentGap->start <= partEnd) && (currentGap->end >= start))
			{
				// Is this the first gap?
				if ((firstGap == NULL) || (currentGap->start < firstGap->start))
					firstGap = currentGap;
			}
		}

		// If no gaps after start, exit
		if (firstGap == NULL)
			return false;

		// Update start position if gap starts after current pos
		if (start < firstGap->start)
			start = firstGap->start;

		// If this is not within part, exit
		if (start > partEnd)
			return false;

		// Find end, keeping within the max block size and the part limit
		end = firstGap->end;
		blockLimit = partStart + (BLOCKSIZE * (((start - partStart) / BLOCKSIZE) + 1)) - 1;
		if (end > blockLimit)
			end = blockLimit;
		if (end > partEnd)
			end = partEnd;
    
		// If this gap has not already been requested, we have found a valid entry
		if (!IsAlreadyRequested(start, end))
		{
			// Was this block to be returned
			if (result != NULL)
			{
				result->StartOffset = start;
				result->EndOffset = end;
				memcpy(result->FileID, GetFileHash(), 16);
				result->transferred = 0;
			}
			return true;
		}
		else
		{
			// Reposition to end of that gap
			start = end + 1;
		}

		// If tried all gaps then break out of the loop
		if (end == partEnd)
			break;
	}

	EMULE_CATCH

	// No suitable gap found
	return false;
}

void CPartFile::FillGap(uint32 start, uint32 end)
{
	EMULE_TRY

	POSITION pos1, pos2;
	for (pos1 = gaplist.GetHeadPosition();(pos2 = pos1) != NULL;){
		Gap_Struct* cur_gap = gaplist.GetNext(pos1);
		if (cur_gap->start >= start && cur_gap->end <= end){ // our part fills this gap completly
			gaplist.RemoveAt(pos2);
			delete cur_gap;
			continue;
		}
		else if (cur_gap->start >= start && cur_gap->start <= end){// a part of this gap is in the part - set limit
			cur_gap->start = end+1;
		}
		else if (cur_gap->end <= end && cur_gap->end >= start){// a part of this gap is in the part - set limit
			cur_gap->end = start-1;
		}
		else if (start >= cur_gap->start && end <= cur_gap->end){
			uint32 buffer = cur_gap->end;
			cur_gap->end = start-1;
			cur_gap = new Gap_Struct;
			cur_gap->start = end+1;
			cur_gap->end = buffer;
			gaplist.InsertAfter(pos1,cur_gap);
			break;	// Lord KiRon
		}
	}

	UpdateCompletedInfos();
	//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	UpdateDisplayedInfo();
	newdate = true;
	// Barry - The met file is now updated in FlushBuffer()
	//SavePartFile();

	EMULE_CATCH
}

void CPartFile::UpdateCompletedInfos() 
{
	EMULE_TRY

   	uint32 allgaps = 0; 

	for (POSITION pos = gaplist.GetHeadPosition();pos !=  0;gaplist.GetNext(pos)){ 
		Gap_Struct* cur_gap = gaplist.GetAt(pos); 
		allgaps += cur_gap->end - cur_gap->start;
	}

	if ((gaplist.GetCount() || requestedblocks_list.GetCount()))
	{ 
			percentcompleted = ((1.0f-(float)allgaps/filesize)) * 100; 
			completedsize = (uint32)((filesize-(allgaps+1))); 
	} else {
		percentcompleted = 100;
		completedsize=filesize;
	}

	EMULE_CATCH
}

void CPartFile::DrawStatusBar(CDC* dc, RECT* rect, bool bFlat)
{
	EMULE_TRY

	COLORREF crProgress; 
	COLORREF crHave; 
	COLORREF crPending = RGB(255, 208, 0);
	COLORREF crMissing = RGB(255, 0, 0); 

	if(bFlat) { 
		crProgress = RGB(0, 150, 0); 
		crHave = RGB(0, 0, 0); 
	} else { 
		crProgress = RGB(0, 192, 0);	//Cax2 - same green used for files & clients
		crHave = RGB(95, 95, 95);		//Cax2 - darker 'black'...
	} 

	s_ChunkBar.SetHeight(rect->bottom - rect->top); 
	s_ChunkBar.SetWidth(rect->right - rect->left); 
	s_ChunkBar.SetFileSize(filesize); 
	s_ChunkBar.Fill(crHave); 

	uint32 allgaps = 0;

	if(status == PS_COMPLETE || status == PS_COMPLETING) {
		s_ChunkBar.FillRange(0, filesize, crProgress); 
		s_ChunkBar.Draw(dc, rect->left, rect->top, bFlat); 
		percentcompleted = 100;
		completedsize=filesize;
		return; 
	} 

	// red gaps 
	for (POSITION pos = gaplist.GetHeadPosition();pos !=  0;gaplist.GetNext(pos)){ 
		Gap_Struct* cur_gap = gaplist.GetAt(pos); 
		allgaps += cur_gap->end - cur_gap->start;
		bool gapdone = false; 
		uint32 gapstart = cur_gap->start; 
		uint32 gapend = cur_gap->end; 
		for (uint32 i = 0; i != GetPartCount(); i++){ 
			if (gapstart >= i*PARTSIZE && gapstart <=  (i+1)*PARTSIZE){ // is in this part? 
				if (gapend <= (i+1)*PARTSIZE) 
				gapdone = true; 
				else{ 
				gapend = (i+1)*PARTSIZE; // and next  part 
				} 
				// paint 
				COLORREF color; 
				if (m_SrcpartFrequency.GetCount() >= (INT_PTR)i && m_SrcpartFrequency[i])  // frequency? 
				color = RGB(0, 
				(210-(22*(m_SrcpartFrequency[i]-1)) <  0)? 0:210-(22*(m_SrcpartFrequency[i]-1)) 
				,255); 
				else 
				color = crMissing; 

				s_ChunkBar.FillRange(gapstart, gapend + 1,  color); 

				if (gapdone) // finished? 
				break; 
				else{ 
				gapstart = gapend; 
				gapend = cur_gap->end; 
				} 
			} 
		} 
	} 

	// yellow pending parts 
	for (POSITION pos = requestedblocks_list.GetHeadPosition();pos !=  0;requestedblocks_list.GetNext(pos))
	{
		Requested_Block_Struct* block =  requestedblocks_list.GetAt(pos); 
		s_ChunkBar.FillRange((block->StartOffset + block->transferred), block->EndOffset,  crPending);
	} 

	s_ChunkBar.Draw(dc, rect->left, rect->top, bFlat); 

	//--- xrmb:confirmeddownload ---
	if((gaplist.GetCount() || requestedblocks_list.GetCount()))
	{
		// all this here should be done in the Process, not in the drawing!!! its a waste of cpu time, or maybe not :)
		uint16	completedParts=0;
		confirmedsize=0;
		for(uint32 i=0; i!=GetPartCount(); i++)
		{
			uint32	end=(i+1)*PARTSIZE-1;
			bool	lastChunk=false;
			//--- last part? ---
			if(end>filesize)
			{
				end=filesize;
				lastChunk=true;
			}

			if(IsComplete(i*PARTSIZE, end))
			{
				completedParts++;

				if(lastChunk==false)
					confirmedsize+=PARTSIZE;
				else
					confirmedsize+=filesize % PARTSIZE;
			}
		}
		
		completedsize = (uint32)((filesize-(allgaps+1)));

		percentcompleted = (float)completedsize/filesize*100;
		percentconfirmed = (float)confirmedsize/filesize*100;
	}
	else 
	{
		percentcompleted = 100;
		percentconfirmed = 100;
		completedsize=filesize;
	}

	// green progress 
	RECT gaprect; 
	gaprect.top = rect->top; 
	gaprect.bottom = gaprect.top + PROGRESS_HEIGHT; // Barry - was 4
	gaprect.left = rect->left; 

	uint32	w=rect->right-rect->left+1;
	uint32	wc=(uint32)(percentconfirmed/100*w+0.5f);
	uint32	wp=(uint32)(percentcompleted/100*w+0.5f);

	if(!bFlat) 
	{
		s_LoadBar.SetWidth(wp);
		s_LoadBar.SetFileSize(completedsize);
		s_LoadBar.Fill(crPending);
		s_LoadBar.FillRange(0, confirmedsize, crProgress);
      	s_LoadBar.Draw(dc, gaprect.left, gaprect.top, false); 
	} 
	else 
	{
		int	left=rect->left;

		// green
		//gaprect.left =  left;
		gaprect.right = left+wc;
      		dc->FillRect(&gaprect, &CBrush(crProgress)); 

		// yellow
		gaprect.left = left+wc;
		gaprect.right = left+wp;
		dc->FillRect(&gaprect, &CBrush(crPending));

 		//draw gray progress only if flat 
		gaprect.left = left+wp;
		gaprect.right = left+w;
      		dc->FillRect(&gaprect, &CBrush(RGB(224,224,224)));
	} 
	//--- :xrmb ---

	EMULE_CATCH
}

void CPartFile::WritePartStatus(CFile* file)
{
	EMULE_TRY

	uint16 parts = GetPartCount();
	file->Write(&parts,2);
	uint16 done = 0;
	while (done != parts){
		uint8 towrite = 0;
		for (uint32 i = 0;i != 8;i++){
			//--- xrmb:partprio ---
			if (GetPartStatus(done)==PR_PART_ON && IsComplete(done*PARTSIZE,((done+1)*PARTSIZE)-1))
			//--- :xrmb ---
				towrite |= (1<<i);

			done++;
			if (done == parts)
				break;
		}
		file->Write(&towrite,1);
	}

	EMULE_CATCH
}

int CPartFile::GetValidSourcesCount() 
{
	EMULE_TRY

	int counter=0;
	POSITION pos1,pos2;
	for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
	for (pos1 = srclists[sl].GetHeadPosition();( pos2 = pos1 ) != NULL;){
		srclists[sl].GetNext(pos1);
		CUpDownClient* cur_src = srclists[sl].GetAt(pos2);
		if (cur_src->GetDownloadState()!=DS_ONQUEUE && cur_src->GetDownloadState()!=DS_DOWNLOADING &&
			cur_src->GetDownloadState()!=DS_NONEEDEDPARTS) counter++;
	}
	return counter;

	EMULE_CATCH

	return 0;
}

uint16 CPartFile::GetNotCurrentSourcesCount()
{
	EMULE_TRY

	uint16 counter=0;

	POSITION pos1,pos2;
		for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
		for (pos1 = srclists[sl].GetHeadPosition();( pos2 = pos1 ) != NULL;){
			srclists[sl].GetNext(pos1);
			CUpDownClient* cur_src = srclists[sl].GetAt(pos2);
		if (cur_src->GetDownloadState()!=DS_ONQUEUE && cur_src->GetDownloadState()!=DS_DOWNLOADING) counter++;
	}
	return counter;

	EMULE_CATCH

	return 0;
}

uint8 CPartFile::GetStatus(bool ignorepause)
{
	EMULE_TRY

	if ((!paused) || status == PS_ERROR || ignorepause)
		return status;
	else
		return PS_PAUSED;

	EMULE_CATCH
}

//Inserted by Vorlost for load rarest parts first, code by Denominator 
uint16 CPartFile::GetRemainingBlocksInPart(uint16 partnumber)
{
	EMULE_TRY

	uint16 result = 0; 
	for (uint32 i = 0; i != ceil((float)PARTSIZE/BLOCKSIZE);i++){ 
		uint32 start = (PARTSIZE*partnumber) + i*BLOCKSIZE; 
		uint32 end    = (PARTSIZE*partnumber) + ((i+1)*BLOCKSIZE)-1; 
		if (end >= PARTSIZE*((uint32)partnumber+1)) 
			end = (PARTSIZE*(partnumber+1))-1; 
		if (start >= GetFileSize()) 
			break; 
		if (end >= GetFileSize()) 
			end = GetFileSize()-1; 
		if ( !IsComplete(start,end) && !IsAlreadyRequested(start,end)) 
			result++; 
	} 
	return result; 

	EMULE_CATCH

	return 0;
} 
//End of code insertion by Vorlost for loading rarest parts first

// Added by Tarod -remove no needed sources from download queue
//
// [In]	number_sources_to_remove: Number of sources to be removed from list or 0xffffffff if we want all removed
void CPartFile::RemoveNoNeededSources(uint32 number_sources_to_remove)
{
	EMULE_TRY

	POSITION pos1, pos2;
	uint32 removed = 0;
    for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
	for (pos1 = srclists[sl].GetHeadPosition(); (pos2 = pos1) != NULL && removed < number_sources_to_remove;) {
		srclists[sl].GetNext(pos1);
		CUpDownClient* cur_src = srclists[sl].GetAt(pos2);
		if (cur_src->GetDownloadState() == DS_NONEEDEDPARTS) 
		{
			theApp.downloadqueue->RemoveSource(cur_src, true);
			removed ++; // Count removed source
		}
	}

	EMULE_CATCH

	lastpurgetime = ::GetTickCount();
}

uint32 CPartFile::Process(uint32 reducedownload/*in percent*/)
{
	EMULE_TRY

	DWORD dwCurTick = ::GetTickCount();

	// If buffer size exceeds limit, or if not written within time limit, flush data
	if ((m_nTotalBufferData > BUFFER_SIZE_LIMIT) /*|| (dwCurTick > (m_nLastBufferFlushTime + BUFFER_TIME_LIMIT))*/)
	{
		// Avoid flushing while copying preview file
		if (!m_bPreviewing)
			FlushBuffer();
	}

	// check if we want new sources from server
	//uint16 test = theApp.glob_prefs->GetMaxSourcePerFileSoft();
	if (( (!lastsearchtime) || (dwCurTick - lastsearchtime) > SERVERREASKTIME) && theApp.serverconnect->IsConnected()
		&& theApp.glob_prefs->GetMaxSourcePerFileSoft() > GetSourceCount() ){
		//local server
		lastsearchtime = dwCurTick;
		Packet* packet = new Packet(OP_GETSOURCES,16);
		memcpy(packet->pBuffer,filehash,16);
		theApp.uploadqueue->AddUpDataOverheadServer(packet->size);
		theApp.serverconnect->SendPacket(packet,true);
		theApp.emuledlg->AddDebugLogLine(false, "Send:Source Request Server File(%s)", GetFileName());
	}

	bool sockets_usage_high = theApp.listensocket->GetOpenSockets() > (theApp.glob_prefs->GetMaxConnections() * 0.75);
	transferingsrc = 0;
	datarate = 0;
	POSITION pos1, pos2;
	for (uint32 sl = 0; sl < SOURCESSLOTS; sl++)
	{
		if (!srclists[sl].IsEmpty())
		{
			for (pos1 = srclists[sl].GetHeadPosition();( pos2 = pos1 ) != NULL;)
			{
		srclists[sl].GetNext(pos1);
		CUpDownClient* cur_src = srclists[sl].GetAt(pos2);

        uint32 usethisfilereasktime = FILEREASKTIME;

        switch (cur_src->GetDownloadState())
		{
		case DS_DOWNLOADING:{
			transferingsrc++;
			uint32 cur_datarate = cur_src->CalculateDownloadRate();
			datarate += cur_datarate;
			if (reducedownload && cur_src->GetDownloadState() == DS_DOWNLOADING){
							uint32 limit = reducedownload*cur_datarate/1000; //(uint32)(((float)reducedownload/100)*cur_datarate)/10;		
					if (limit < 1000 && reducedownload == 200)
						limit += 1000;
					else if (limit < 1)
						limit = 1;
					cur_src->socket->SetDownloadLimit(limit);
			}
			else if( m_nSavedReduceDownload != reducedownload )
				cur_src->socket->DisableDownloadLimit();
			m_nSavedReduceDownload  = reducedownload;

			break;
		}
		case DS_BANNED:
		case DS_ERROR:
			break;
		case DS_LOWTOLOWIP:	// if we now have a high ip we can ask
			if( ((dwCurTick - lastpurgetime) > PURGE_TIME) && (GetSourceCount() >= (theApp.glob_prefs->GetMaxSourcePerFile()*.8 )) ){
				theApp.downloadqueue->RemoveSource( cur_src );
				lastpurgetime = dwCurTick;
							break;
			}
			if (theApp.serverconnect->IsLowID())
				break;
		case DS_NONEEDEDPARTS:{ 

			// doubled reasktime for no needed parts - save connections and traffic
			//Vorlost, noneededparts-sources will be checked filereasktime*4 when 3/4 of max. sockets are open 
			if (sockets_usage_high)
				usethisfilereasktime = FILEREASKTIME * 2;
			//End Vorlost for Expanded Filereasktime 

			// doubled reasktime for no needed parts - save connections and traffic
			if (!((!cur_src->GetLastAskedTime()) || (dwCurTick - cur_src->GetLastAskedTime()) > usethisfilereasktime*2))
				break;
		}
		case DS_ONQUEUE:{
	        if ( cur_src->GetRemoteQueueRank()
		        && sockets_usage_high
		        ) { //Vorlost QueueRank > 250 --> Filereasktime=45minutes
		        if ( cur_src->GetRemoteQueueRank() > 250 )
			        usethisfilereasktime = FILEREASKTIME * 2.5 ; //Vorlost
		        else usethisfilereasktime = FILEREASKTIME * 2; //Vorlost (QR 1-250 = 36minutes)

	        }

            if (theApp.serverconnect->IsConnected() && ((!cur_src->GetLastAskedTime()) || (dwCurTick - cur_src->GetLastAskedTime()) > usethisfilereasktime-20000)){
				cur_src->UDPReaskForDownload();
			}

			// Check Remote queue position
			if ((!cur_src->IsRemoteQueueFull()) && (cur_src->GetRemoteQueueRank() > theApp.glob_prefs->GetRemoteQueueLimit())) {
				theApp.downloadqueue->RemoveSource(cur_src);
                break;
			}
		}
		case DS_CONNECTING:
		case DS_TOOMANYCONNS:
		case DS_CONNECTED:
		case DS_NONE:
		case DS_WAITCALLBACK:
			//Vorlost for Expanded FILEREASKTIME if QR is known and 3/4 Sockets are open 
			if (theApp.serverconnect->IsConnected() && ((!cur_src->GetLastAskedTime()) || (dwCurTick - cur_src->GetLastAskedTime()) > usethisfilereasktime))
				cur_src->AskForDownload();
			break;
		}
	}
        }
    }


	// swap No needed partfiles if possible
	// Modified by Tarod -now handles A4AF too keeping usage of CPU less and do not forcing files
	// that has No Needed Parts to download letting them swap to another file
	if (((!m_LastNoNeededCheck) || (dwCurTick - m_LastNoNeededCheck) > 10000)){
		m_LastNoNeededCheck = dwCurTick;
		POSITION pos1,pos2;
		for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
		for (pos1 = srclists[sl].GetHeadPosition();( pos2 = pos1 ) != NULL;){
			srclists[sl].GetNext(pos1);
			CUpDownClient* cur_src = srclists[sl].GetAt(pos2);
			if (cur_src->GetDownloadState() == DS_NONEEDEDPARTS)
				cur_src->SwapToAnotherFile(false);
			// if this file has activated A4AF auto just hang to another file if we can
			else if (IsA4AFAuto() && (cur_src->GetDownloadDatarate() == DS_ONQUEUE || cur_src->GetDownloadDatarate() == DS_LOWTOLOWIP)
				&& memcmp(cur_src->reqfileid, GetFileHash(), 16) != 0) {

					cur_src->HandoverToAnotherFile(this);
			}
		}

		// Added by Tarod -if is A4AF auto then swith every A4AF source to this file
		if (IsA4AFAuto()) {
			theApp.emuledlg->transferwnd.downloadlistctrl.DownloadAllA4AF(this);
		}
	}

/*
	// Added by Tarod (remove no needed parts after FILEREASKTIME * 2 + 10000)
	if (theApp.glob_prefs->GetMaxSourcePerFileSoft() == GetSourceCount()
		&& theApp.glob_prefs->IsRemoveNNSEnabled() //Vorlost
	    && ((m_LastRemovedTime == 0) || (::GetTickCount() - m_LastRemovedTime) > FILEREASKTIME * 2 + 10000))
	{
		m_LastRemovedTime = ::GetTickCount();
		RemoveNoNeededSources();
	}
	else if (theApp.glob_prefs->IsRemoveNNSEnabled() 
	         && ((m_LastRemovedTime == 0) || (::GetTickCount() - m_LastRemovedTime) > FILEREASKTIME * 2 + 10000))
	{
		m_LastRemovedTime = ::GetTickCount();
		RemoveNoNeededSources();
	}
*/

	// calculate datarate, set limit etc.
	count++;
	if (count == 30){
		count = 0;
		//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
		UpdateDisplayedInfo();
		if(m_bPercentUpdated == false)
			UpdateCompletedInfos();
		m_bPercentUpdated = false;
	}
	///////////////////////////////////////////////////////
	//for saving/reloading source to/from disk: mr fry
#ifdef OLD_SLS_ENABLED
	if(GetStatus() < 2 && theApp.downloadqueue->IsSourcesLoadingEnabled()){
		uint32 currentTime = ::GetTickCount();

		if(!m_bWasConnectedToServerLastTime && theApp.serverconnect->IsConnected()){
			//if connection-state has switch from: not connected -> connected...
			TRACE("CPartFile::Process : slow loading of sources down if a connection is established (next time sources are loaded for %s: in %.1f sec)\n"
				,GetFileName(),(float(m_iLoadingSourcesPartialTimeInterval)/1000));
			m_iLastTimeSourcesLoadedPartial = ::GetTickCount();
		}

		//load sources in slow mode (while they weren't completely loaded yet)?
		if(!m_bLoadedSourcesCompletely && theApp.serverconnect->IsConnected() && !theApp.downloadqueue->IsMaxClientsExceeded())
//			&& theApp.listensocket->GetOpenSockets() < theApp.glob_prefs->GetMaxConnections()-20-RESTORE_SOURCES_AT_ONCE) // Vorlost
		{  
			if(currentTime - m_iLastTimeSourcesLoadedPartial > m_iLoadingSourcesPartialTimeInterval)
			{
				// makes room for more loaded sources
				if (theApp.glob_prefs->IsRemoveNNSEnabled()) //Vorlost added prefs question
				{
					int source_count_plus_needed = GetSourceCount() + GetLoadSourcesAtOnceLimit();
					if (source_count_plus_needed > theApp.glob_prefs->GetMaxSourcePerFile()
					&& ((lastpurgetime == 0) || (::GetTickCount() - lastpurgetime) > PURGE_TIME))
						RemoveNoNeededSources(source_count_plus_needed - theApp.glob_prefs->GetMaxSourcePerFile());
				}
	
				//is a permanent opened CStoredSources-object for each part-file wanted?
				if(theApp.downloadqueue->KeepStoredSourcesOpen())
					AddFromStoredSources(m_stscrStoringSources,false,GetLoadSourcesAtOnceLimit());
				else
					LoadAndAddSources(false,false,GetLoadSourcesAtOnceLimit());
			}
		}
		//(re-) load sources?
		else if((currentTime - m_iLastTimeSourcesLoaded) > m_iLoadingSourcesTimeInterval  
				&& theApp.serverconnect->IsConnected()
//				&& theApp.listensocket->GetOpenSockets() < theApp.glob_prefs->GetMaxConnections()-20-RESTORE_SOURCES_AT_ONCE) // Vorlost
				&& !theApp.downloadqueue->IsMaxClientsExceeded()) // from 25b code. don't know if it's better
		{
			TRACE("Reloading all sources for %s\n",GetFileName());

			// Added by Tarod -before reload purge outdated and no needed sources
			// remember that this only occurs first time -first group of loaded sources
			if(m_stscrStoringSources != NULL)
			{
				m_stscrStoringSources->DeleteOutdatedSources();
				m_stscrStoringSources->SortSourcesByPartsNDate();
				m_stscrStoringSources->ReduceSourcesTo(m_stscrStoringSources->GetMaxSourceEntries());
			}
			if (theApp.glob_prefs->IsRemoveNNSEnabled())  //Vorlost added prefs question
			{
			    int source_count_plus_needed = GetSourceCount() + GetLoadSourcesAtOnceLimit();
			    if (source_count_plus_needed > theApp.glob_prefs->GetMaxSourcePerFile())
				    RemoveNoNeededSources(source_count_plus_needed - theApp.glob_prefs->GetMaxSourcePerFile());
			}
			
			//is a permanent opened CStoredSources-object for each part-file wanted?
			if(theApp.downloadqueue->KeepStoredSourcesOpen()) {
				//as in the previous if(), but with AddFrom..(..,true); -> reset=true means, begin again to load all sources from stored-sources
				AddFromStoredSources(m_stscrStoringSources,false,GetLoadSourcesAtOnceLimit(),true);
			} else
				LoadAndAddSources(false,false,GetLoadSourcesAtOnceLimit(),true);
		}
		//save sources?
		if((currentTime - m_iLastTimeSourcesSaved) > m_iSavingSourcesTimeInterval
			&& m_bLoadedSourcesCompletely){

			//only save to disk now, if there is no permantent opened CStoredSources-object for this part-file
			if(!theApp.downloadqueue->KeepStoredSourcesOpen()){
				TRACE("(automatically) saving sources for %s to disk\n",GetFileName());
				SaveSources(true);
			}
			else if(m_stscrStoringSources != NULL){//sort, so we get the best sources first, the next time they're loaded
				//if (!m_stscrStoringSources->IsRareFile()) m_stscrStoringSources->DeleteOutdatedSources();
				m_stscrStoringSources->SortSourcesByPartsNDate();
				SaveSources(false);
				//m_stscrStoringSources->ReduceSourcesTo(m_stscrStoringSources->GetMaxSourceEntries());
			}

			// Added by Tarod fixed bug of CPU usage (por fin, me cago en la puta hostia, joder!!!)
			m_iLastTimeSourcesSaved = ::GetTickCount();
		}
        
		m_bWasConnectedToServerLastTime = theApp.serverconnect->IsConnected();
	}//end: if(status)
	////////////////////////////////////////////////////////
#endif //OLD_SLS_ENABLED
	return datarate;

	EMULE_CATCH

	return 0;
}

void CPartFile::AddSources(CMemFile* sources,uint32 serverip, uint16 serverport)
{
	EMULE_TRY
	
	//if this part file is paused or complete/completing no sources needed any more
	if(GetStatus()==PS_PAUSED || GetStatus() > 7)		//Cax2 + SyruS check
		return;

	uint8 count;
	uint8 debug_lowiddropped = 0;
	uint8 debug_possiblesources = 0;
	sources->Read(&count,1);
	if (theApp.glob_prefs->IsRemoveNNSEnabled())
    {
	    // Added by Tarod - if there is needed place for more sources just remove NNS
	    int source_count_plus_new = GetSourceCount() + count;
	    if (source_count_plus_new > theApp.glob_prefs->GetMaxSourcePerFile()
		    && ((lastpurgetime == 0) || (::GetTickCount() - lastpurgetime) > PURGE_TIME))
		    RemoveNoNeededSources(source_count_plus_new - theApp.glob_prefs->GetMaxSourcePerFile());
    }
	for (int i = 0;i != count;i++){
		uint32 userid;
		sources->Read(&userid,4);
		uint16 port;
		sources->Read(&port,2);
		// check first if we are this source

		// MOD Note: Do not change this part - Merkur
		if (theApp.serverconnect->GetClientID() < 16777216 && theApp.serverconnect->IsConnected()){
			if ((theApp.serverconnect->GetClientID() == userid) && inet_addr(theApp.serverconnect->GetCurrentServer()->GetFullIP()) == serverip)
				continue;
		}
		else if (theApp.serverconnect->GetClientID() == userid)
			continue;
		else if (userid < 16777216 && !theApp.serverconnect->IsLocalServer(serverip,serverport)){
			debug_lowiddropped++;
			continue;
		}
		// MOD Note - end

		if( theApp.glob_prefs->GetMaxSourcePerFile() > GetSourceCount() ){
			debug_possiblesources++;
			CUpDownClient* newsource = new CUpDownClient(port,userid,serverip,serverport,this);
			theApp.downloadqueue->CheckAndAddSource(this,newsource);
		} 
	}
//	theApp.emuledlg->AddDebugLogLine(false,"RCV: %i sources from server, %i low id dropped, %i possible sources",count,debug_lowiddropped,debug_possiblesources);

	EMULE_CATCH
}

void CPartFile::NewSrcPartsInfo()
{
	EMULE_TRY

    uint16 partcount;

    partcount = GetPartCount();

	if (m_SrcpartFrequency.GetCount() != partcount){
		m_SrcpartFrequency.RemoveAll();
		m_SrcpartFrequency.SetSize(0);
		for (int i = 0; i != partcount;i++)
			m_SrcpartFrequency.Add(0);
	}
	else{
		for (int i = 0; i != partcount;i++)
			m_SrcpartFrequency[i] = 0;
	}
	for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
	for (POSITION pos = srclists[sl].GetHeadPosition();pos != 0;){
        CUpDownClient* cur_src = srclists[sl].GetNext(pos);
        uint8* partstatus = cur_src->GetPartStatus();
        if(partstatus)
		    for (int i = 0; i != partcount; i++){
			    if (partstatus[i])
				m_SrcpartFrequency[i] +=1;
		}
	}
	//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	UpdateDisplayedInfo();

	EMULE_CATCH
}

//Vorlost for load rarest chunks first patch2
bool CPartFile::GetNextRequestedBlock(CUpDownClient* sender,Requested_Block_Struct** newblocks,uint16* count)
{ 
	EMULE_TRY

	if(sender == NULL || newblocks == NULL || count == NULL)
		return false;

//	#define BLOCKCOUNT 3 

	uint16 requestedCount = *count;
	uint16 newblockcount = 0; 
	uint8* partsav = sender->GetPartStatus(); 

  	uint8 mode = GetMovieMode();
  	// Mode 0=NoPreview,1=AVI(1F,1L),2=AVI(2F,1L),3=AVI(3F,1L)
  	// Mode 4=MPG(1F),5=MPG(2F),6=MPG(3F),7=AVI/MPG(2F,2L) F=First,L=Last
  	uint8 rarechunk = 10; //Up to this value a chunk is very rare,if more we complete first
 	uint8 raremovie = 5; //Up to this value no moviechunks will be loaded first
    //, but could happen earlier if it doesnt make any difference(preview enabled), otherwise random
  	bool forcepreview = false; //true will ignore rare-chunks completely if a moviechunk is available

	*count = 0; 
    uint16 randomness;
    CList<int,int> liGoodParts;
    CList<int,int> liPossibleParts;
	bool finished = false; 
	while (!finished)
	{ 
/*
		uint16 randomness = (uint16)ROUND(((float)rand()/RAND_MAX)*(GetPartCount()-1)); 
		uint16 goodpart = 0xFFFF; 
		for (uint16 i = 0;i != GetPartCount();i++)
		{ 
			if (partsav[randomness] && !IsComplete(randomness*PARTSIZE,((randomness+1)*PARTSIZE)-1))
			{ 
				// Begin of absolutely lowest availability gap by Vorlost inspired by Denominator 
           		if (GetNextEmptyBlockInPart(randomness,0)) //Search for next empty Block... 
				{
					if (goodpart != 0xFFFF) //One Part memorized yet ? 
					{
						if (m_SrcpartFrequency[goodpart] > rarechunk && m_SrcpartFrequency[randomness] > rarechunk) //Non-rare chunks found yet
						{
                    		if (PrepareComparePart(randomness) < PrepareComparePart(goodpart)) //and is this new chunk more completed ?
							{
     							if (Movie1(mode,goodpart,randomness)) goodpart = randomness;  //Then we want it ! 
                    		} 
     						else
							{
      							if (PrepareComparePart(randomness) == PrepareComparePart(goodpart) //Is the new chunk same complete
      								&& m_SrcpartFrequency[randomness] < m_SrcpartFrequency[goodpart]) //and rarer ?
								{
      									if (Movie1(mode,goodpart,randomness))
											goodpart = randomness;  //Then we want it ! If Movie1 allows us
								} 
     							else if (Movie2(mode,raremovie,forcepreview,goodpart,randomness))
									goodpart = randomness; //Normally we dont want it, what do you think Movie2 ?
							} 
						} 
               			else //The memorized or the new chunk is a very rare one
						{
                   			if (m_SrcpartFrequency[randomness] < m_SrcpartFrequency[goodpart]) //Is the new chunk rarer than the memorized ? 
								goodpart = randomness; //Then we want it ! 
                   			else if (m_SrcpartFrequency[randomness] == m_SrcpartFrequency[goodpart]) //old & new chunk are both same available ?
							{
                   				if (PrepareComparePart(randomness) < PrepareComparePart(goodpart)) //and is this new chunk more completed ?
								{
                       				if (Movie1(mode,goodpart,randomness))
										goodpart = randomness; //Then we want it ! If Movie1 allows us
								} 
  								else if (Movie2(mode,raremovie,forcepreview,goodpart,randomness))
									goodpart = randomness; //Normally we dont want it, what do you think Movie2 ?
							} 
						} 
					} else	                      //Theres no memorized Part yet... 
						goodpart = randomness;    //Then just take this one first...    
				} 
				// End of lowest availability gap by Vorlost / Denominator 
			} 
			randomness++; 
			if (randomness == GetPartCount()) 
				randomness = 0; 
		} 
		uint16 usedpart; 
		if (goodpart != 0xFFFF)
			usedpart = goodpart; 
		else
		{ 
			if (!newblockcount) 
				return false;
			else 
				break; 
		} 
		// Ende von Vorlosts Code 
*/
                // Need to empty lists to avoid infinite loop when file is smaller than 180k
                // Otherwise it would keep looping to find 3 blocks, there is only one and it is requested
                liGoodParts.RemoveAll();
                liPossibleParts.RemoveAll();
                // Barry - Top priority should be to continue downloading from current blocks (if anything left to download)
                bool foundPriorityPart = false;
                CString gettingParts;
                sender->ShowDownloadingParts(&gettingParts);
                
                for (int i=(gettingParts.GetLength()-1); i>=0; i--)
                {
                        if ((gettingParts.GetAt(i) == 'Y') && (GetNextEmptyBlockInPart(i, 0)))
                        {
                                liGoodParts.AddHead(i);
                                foundPriorityPart = true;
                        }
                }
                // Barry - Give priorty to end parts of archives and movies
                if ((!foundPriorityPart) && (IsArchive() || IsMovie()) && (GetMovieMode()>0)) // obaldin
                {
                        uint32 partCount = GetPartCount();
                        // First part
                        if (sender->IsPartAvailable(0) && GetNextEmptyBlockInPart(0, 0))
                        {
                                liGoodParts.AddHead(0);
                                foundPriorityPart = true;
                        }
                        else if ((partCount > 1))
                        {
                                // Last part
                                if (sender->IsPartAvailable(partCount-1) && GetNextEmptyBlockInPart(partCount-1, 0))
                                {
                                        liGoodParts.AddHead(partCount-1);
                                        foundPriorityPart = true;
                                }
                                // Barry - Better to get rarest than these, add to list, but do not exclude all others.
                                // These get priority over other parts with same availability.
                                else if (partCount > 2)
                                {
                                        // Second part
                                        if (sender->IsPartAvailable(1) && GetNextEmptyBlockInPart(1, 0))
                                                liGoodParts.AddHead(1);
                                        // Penultimate part
                                        else if (sender->IsPartAvailable(partCount-2) && GetNextEmptyBlockInPart(partCount-2, 0))
                                                liGoodParts.AddHead(partCount-2);
                                }
                        }
                }
                if (!foundPriorityPart)
                {
                        randomness = (uint16)ROUND(((float)rand()/RAND_MAX)*(GetPartCount()-1));
                        for (uint16 i = 0;i != GetPartCount();i++){
                                if (sender->IsPartAvailable(randomness))
                                {                                        
										if (partsav[randomness] && !IsComplete(randomness*PARTSIZE,((randomness+1)*PARTSIZE)-1)){
                                                /*if (IsCorruptedPart(randomness)){
                                                        if (GetNextEmptyBlockInPart(randomness,0)){
                                                                goodpart = randomness;
                                                                break;
                                                        }
                                                }
                                                else */ 
                                                if (IsPureGap(randomness*PARTSIZE,((randomness+1)*PARTSIZE)-1))
                                                {
                                                        if (GetNextEmptyBlockInPart(randomness,0))
                                                                liPossibleParts.AddHead(randomness);
                                                }
                                                else if (GetNextEmptyBlockInPart(randomness,0))
                                                        liGoodParts.AddTail(randomness); // Barry - Add after archive/movie entries
                                        }
                                }
                                randomness++;
                                if (randomness == GetPartCount())
                                        randomness = 0;
                        }
                }
                CList<int,int>* usedlist;
                if (!liGoodParts.IsEmpty())
                        usedlist = &liGoodParts;
                else if (!liPossibleParts.IsEmpty())
                        usedlist = &liPossibleParts;
                else{
                        if (!newblockcount){
                                return false;
                        }
                        else
                                break;
                }
                uint16 nRarest = 0xFFFF;
                uint16 usedpart = usedlist->GetHead();
                for (POSITION pos = usedlist->GetHeadPosition();pos != 0;usedlist->GetNext(pos)){
                        if (m_SrcpartFrequency.GetCount() >= usedlist->GetAt(pos)
                        && m_SrcpartFrequency[usedlist->GetAt(pos)] < nRarest){
                                nRarest = m_SrcpartFrequency[usedlist->GetAt(pos)];
                                usedpart = usedlist->GetAt(pos);
                        }
                }
		while (true)
		{ 
			Requested_Block_Struct* block = new Requested_Block_Struct; 
			if (GetNextEmptyBlockInPart(usedpart,block))
			{ 
				requestedblocks_list.AddTail(block); 
				newblocks[newblockcount] = block; 
				newblockcount++; 
				if (newblockcount == requestedCount)	//BLOCKCOUNT
				{ 
					finished = true; 
					break; 
				} 
			} 
			else 
			{ 
				delete block; 
				break; 
			} 
		} 
	} //wend 
	*count = newblockcount; 
	return true;

	EMULE_CATCH

	return false;
}
//End vorlost

void  CPartFile::RemoveBlockFromList(uint32 start,uint32 end)
{
	EMULE_TRY

	POSITION pos1,pos2;
	for (pos1 = requestedblocks_list.GetHeadPosition(); ( pos2 = pos1 ) != NULL; ){
	   requestedblocks_list.GetNext(pos1);
		if (requestedblocks_list.GetAt(pos2)->StartOffset <= start && requestedblocks_list.GetAt(pos2)->EndOffset >= end)
			requestedblocks_list.RemoveAt(pos2);
	}

	EMULE_CATCH
}

void CPartFile::RemoveAllRequestedBlocks(void)
{
	requestedblocks_list.RemoveAll();
}

void CPartFile::CompleteFile(bool bIsHashingDone)
{
	EMULE_TRY

	if( srcarevisible )
		theApp.emuledlg->transferwnd.downloadlistctrl.HideSources(this);
	status = PS_COMPLETING;
	if (!bIsHashingDone){
		datarate = 0;
		char* partfileb = nstrdup(partmetfilename);
		partfileb[strlen(partmetfilename)-4] = 0;
		CAddFileThread* addfilethread = (CAddFileThread*) AfxBeginThread(RUNTIME_CLASS(CAddFileThread), THREAD_PRIORITY_BELOW_NORMAL,0, CREATE_SUSPENDED);
		addfilethread->SetValues(0,tempdir,partfileb,this);
		addfilethread->ResumeThread();	
		delete[] partfileb;
		return;
	}
	else {
		StopFile();
		CWinThread *pThread = AfxBeginThread((AFX_THREADPROC )CompleteThreadProc,this, THREAD_PRIORITY_BELOW_NORMAL,0, CREATE_SUSPENDED);
		if (pThread)
		{
			pThread->ResumeThread();
		} else {
			throw CString(_T("Error creating file complete thread")); // move this to resources
		}
	} 

	UpdateDisplayedInfo(true);
	EMULE_CATCH
}

// Lord KiRon - using threads for file completion
BOOL CPartFile::PerformFileComplete(void) 
{ 
	EMULE_TRY

	CSingleLock(&m_FileCompleteMutex,TRUE); // will be unlocked on exit
	char* partfilename = nstrdup(fullname);
	partfilename[strlen(fullname)-4] = 0;
	char* newfilename = nstrdup(GetFileName());
	strcpy(newfilename, (LPCTSTR)theApp.StripInvalidFilenameChars(newfilename));
	char* newname = new char[strlen(newfilename)+strlen(theApp.glob_prefs->GetIncomingDir())+MAX_PATH];
	sprintf(newname,"%s\\%s",theApp.glob_prefs->GetIncomingDir(),newfilename);
	
	// close permanent handle
	if (m_hpartfile.m_hFile != INVALID_HANDLE_VALUE)
		m_hpartfile.Close(); // m_hpartfile isn't the handle to the file...
	
	bool renamed = false;
	if(PathFileExists(newname))
	{
		renamed = true;
		int namecount = 0;
		
		size_t length = strlen(newfilename);
		ASSERT(length != 0); //name should never be 0
		
		//the file extension
		char *ext = strrchr(newfilename, '.');
		if(ext == NULL)
			ext = newfilename + length;
		
		char *last = ext;  //new end is the file name before extension
		last[0] = 0;  //truncate file name
		
		//search for matching ()s and check if it contains a number
		if((ext != newfilename) && (strrchr(newfilename, ')') + 1 == last))
		{
			char *first = strrchr(newfilename, '('); 
            
            if(first != NULL)
            {
            	first++;
            	bool found = true;
            	for(char *step = first; step < last - 1; step++)
            		if(*step < '0' || *step > '9')
            		{
            			found = false;
            			break;
            		}
            	if(found)
            	{
            		namecount = atoi(first);
            		last = first - 1;
            		last[0] = 0;  //truncate again
            	}
            }
		} 

		CString strTestName;
		do
		{ 
        	namecount++; 
            strTestName.Format("%s\\%s(%d).%s", theApp.glob_prefs->GetIncomingDir(), 
               				   newfilename, namecount, min(ext + 1, newfilename + length)); 
        }
        while(PathFileExists(strTestName));
        	delete[] newname;
        newname = nstrdup(strTestName);
	} 
    delete newfilename; 

	// Lord KiRon
	BOOL bSlowComplete   =   theApp.glob_prefs->SlowComplete();
	int  BlockSize   =   theApp.glob_prefs->SlowCompleteBlockSize()*1024; 
       
    // we do this only if SlowComplete enabled 
    // AND the files are on the different HDs (paratitions) - check by the drive leter 
    // AND any of path is not in URL/Network format (has : as second leter) 
    if (bSlowComplete && (_toupper(partfilename[0]) != _toupper(newname[0]) && partfilename[1] == _T(':') 
         && newname[1] == _T(':'))) 
    {
    	CFile fInputFile; 
        CFile fOutputFile; 
        try 
        { 
      		CString strTempNew   =   CString(newname); 
            strTempNew += _T(".$$$"); 
            // We going to write everithing into the temporary file at same place the newname located 
            // and after we finished - rename , this will take care of syncronization problem with SharedFiles 
            if (!fInputFile.Open(partfilename,CFile::modeRead | CFile::shareDenyNone | CFile::typeBinary 
                | CFile::osSequentialScan)) 
            { 
               	// let catch to care about closing everything 
               	throw CString(_T(GetResString(IDS_ERR_SLOWREAD)+" ("+partfilename+")"));
               	return FALSE; 
            } 
            if (!fOutputFile.Open(strTempNew,CFile::modeWrite | CFile::shareExclusive | CFile::osSequentialScan | CFile::modeCreate)) 
            { 
            	fInputFile.Close(); 
            	// let catch to care about closing everything 
				throw CString(_T(GetResString(IDS_ERR_SLOWWRITE)+" ("+strTempNew+")"));
               	return FALSE; 
            } 

            DWORD dwRead = 0; 
            char *buffer = new char[BlockSize]; 
            if (!buffer) 
            { 
               	// let catch to care about closing everything 
               	throw CString(_T(GetResString(IDS_ERR_OUTMEM))); 
               	return FALSE; 
            } 

            // Most of the work done here 
            // just copy the file data from one to another 
            do 
            { 
            	dwRead = fInputFile.Read(buffer, BlockSize ); 
               	fOutputFile.Write(buffer, dwRead); 
               	Sleep(0); // release the CPU for other processes 
               	// may be real sleep needed here , not sure 
            } 
            while (dwRead > 0); 
            delete []buffer; 

            fInputFile.Close(); 
            fOutputFile.Close(); 
            // Rename temporary file back to normal name 
            if (rename(strTempNew,newname))
            { 
            	// let catch to care about closing everything 
               	throw CString(_T(GetResString(IDS_ERR_RENFAIL))); 
               	return FALSE; 
            }
            CFile::Remove(partfilename); // This removes old Part file
		} 
        catch(CString str) 
        { 
         	delete[] partfilename; 
            delete[] newname; 
            theApp.emuledlg->AddLogLine(true,str); 
            paused = true; 
            status = PS_ERROR; 
            return FALSE; 
        } 
        catch(...) 
        { 
            delete[] partfilename; 
            delete[] newname; 
            theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_COMPLETIONFAILED),GetFileName()); 
            paused = true; 
            status = PS_ERROR; 
            return FALSE; 
        } 
	} else {
		if (rename(partfilename,newname))
		{ 
        	delete[] partfilename; 
            delete[] newname; 
            theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_COMPLETIONFAILED),GetFileName()); 
            paused = true; 
            status = PS_ERROR; 
            return FALSE; 
        }
	} 
    // Lord KiRon - end
       
    if (remove(fullname)) 
    	theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_DELETEFAILED),fullname); 

    //////////////////////////////// 
    //remove *.src file (thx to brotbuexe) 
 	CString srcName="";
 	srcName.Format("%s\\%s.stsrc",tempdir,GetPartMetFileName());
 	if (remove(srcName))
 	if (errno != ENOENT)
 		theApp.emuledlg->AddLogLine(true,GetResString(IDS_FAILDELETE),srcName); 
    
	// Tarod - erase .backup file
	srcName.Format("%s\\%s.bak", theApp.glob_prefs->GetTempDir(), this->GetPartMetFileName());
	if (remove(srcName))
		if (errno != ENOENT)
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_FAILDELETE),srcName);

	// DonGato - remove .movie file
	srcName.Format("%s\\%s.movie", theApp.glob_prefs->GetTempDir(), this->GetPartMetFileName());
	if (remove(srcName))
		if (errno != ENOENT)
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_FAILDELETE),srcName);

	//--- xrmb:keepPartFileStats ---
	//--- remove the stats file ---
	RemovePartFileStats();
	//--- :xrmb ---

    delete[] partfilename; 
    delete[] fullname; 
    fullname = newname; 
    delete[] directory; 
    directory = nstrdup(theApp.glob_prefs->GetIncomingDir()); 
    status = PS_COMPLETE; 
    paused = false; 
    theApp.emuledlg->AddLogLine(true,GetResString(IDS_DOWNLOADDONE),GetFileName()); 
    theApp.emuledlg->ShowNotifier(GetResString(IDS_TBN_DOWNLOADDONE)+"\n"+GetFileName(), TBN_DLOAD); 
    if (renamed) 
    	theApp.emuledlg->AddLogLine(true, GetResString(IDS_DOWNLOADRENAMED), strrchr(newname, '\\') + 1); 
    theApp.knownfiles->SafeAddKFile(this); 

    // Added to change UAP to Normal after finished download of file by Tarod 
    if (CKnownFile::IsAutoPrioritized())
    { 
    	CKnownFile::SetAutoPriority(false); 
        CKnownFile::SetPriority(PR_NORMAL); 
    } 
    // End Tarod 

    theApp.downloadqueue->RemoveFile(this); 
    //theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this); 
		UpdateDisplayedInfo();
	//SHAddToRecentDocs(SHARD_PATH, fullname); // zegzav (added)

    // Added by Tarod (starts next paused file) 
    if (theApp.glob_prefs->DownloadPausedOnComplete())
    { 
    	theApp.downloadqueue->StartPausedDownload() ; 
    } 
	// Barry - Just in case
//	transfered = filesize;
   	return TRUE; 

	EMULE_CATCH

   	return FALSE; 
} 


void  CPartFile::RemoveAllSources(bool bTryToSwap)
{
	EMULE_TRY

	//TODO transfer sources to other downloading files if possible
	POSITION pos1,pos2;
	for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
	for( pos1 = srclists[sl].GetHeadPosition(); ( pos2 = pos1 ) != NULL; ){
		srclists[sl].GetNext(pos1);
		if (bTryToSwap){
			if (!srclists[sl].GetAt(pos2)->SwapToAnotherFile(true))
				theApp.downloadqueue->RemoveSource(srclists[sl].GetAt(pos2));
		}
		else
			theApp.downloadqueue->RemoveSource(srclists[sl].GetAt(pos2));
	}

	UpdateFileRatingCommentAvail();
	EMULE_CATCH
}

void CPartFile::DeleteFile()
{
	EMULE_TRY

	ASSERT ( !m_bPreviewing );

	StopFile();		//Cax2 bugfix: hanging sources when deleting file
	theApp.sharedfiles->RemoveFile(this);
	theApp.downloadqueue->RemoveFile(this);
	theApp.emuledlg->transferwnd.downloadlistctrl.RemoveFile(this);
	//RemoveAllSources(true);		//Cax2 bugfix: hanging sources when deleting file

	if (m_hpartfile.m_hFile != INVALID_HANDLE_VALUE) m_hpartfile.Close(); // m_hpartfile isn't the handle to the file...

	if (remove(fullname))
		theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_DELETE),fullname);
	char* partfilename = nstrdup(fullname);
	partfilename[strlen(fullname)-4] = 0;
	if (remove(partfilename))
		theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_DELETE),partfilename);

	// remove *.src file (thx to brotbuexe)
	CString srcName="";
	srcName.Format("%s\\%s.stsrc",tempdir,GetPartMetFileName());
	if (remove(srcName))
		if (errno != ENOENT) 
			theApp.emuledlg->AddLogLine(true,GetResString(IDS_FAILDELETE),srcName);

	// Tarod - erase .backup file
	srcName.Format("%s\\%s.bak", theApp.glob_prefs->GetTempDir(), this->GetPartMetFileName());
	if (remove(srcName))
		if (errno != ENOENT)
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_FAILDELETE),srcName);

	// DonGato - remove .movie file
	srcName.Format("%s\\%s.movie", theApp.glob_prefs->GetTempDir(), this->GetPartMetFileName());
	if (remove(srcName))
		if (errno != ENOENT)
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_FAILDELETE),srcName);

	//--- xrmb:keepPartFileStats ---
	//--- remove the stats file ---
	RemovePartFileStats();
	//--- :xrmb ---

	delete[] partfilename;
	delete this;

	EMULE_CATCH
}

bool CPartFile::HashSinglePart(uint16 partnumber)
{
	EMULE_TRY

	if ((GetHashCount() <= partnumber) && (GetPartCount() > 1)){
		theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_HASHERRORWARNING),GetFileName());
		this->hashsetneeded = true;
		return true;
	}
	else if(!GetPartHash(partnumber) && GetPartCount() != 1){
		theApp.emuledlg->AddLogLine(true,GetResString(IDS_ERR_INCOMPLETEHASH),GetFileName());
		this->hashsetneeded = true;
		return true;		
	}
	else{
		uchar hashresult[16];
		m_hpartfile.Seek(PARTSIZE*partnumber,0);
		uint32 length = PARTSIZE;
		if (PARTSIZE*(partnumber+1) > m_hpartfile.GetLength())
			length = (m_hpartfile.GetLength()- (PARTSIZE*partnumber));
		CreateHashFromFile(&m_hpartfile,length,hashresult);

		if (GetPartCount() > 1){
			if (memcmp(hashresult,GetPartHash(partnumber),16))
				return false;
			else
				return true;
		}
		else{
			if (memcmp(hashresult,filehash,16))
					return false;
				else
					return true;
		}
	}

	EMULE_CATCH

	return false;
}

bool CPartFile::IsCorruptedPart(uint16 partnumber)
{
	EMULE_TRY

	return corrupted_list.Find(partnumber);

	EMULE_CATCH

	return true;
}

bool CPartFile::IsMovie()
{
	EMULE_TRY

	// DonGato
	CString fname = GetFileName();
	fname.MakeLower();

	if(IsVLCInstalled())
	{
		return (strstr(fname,".avi") || strstr(fname,".bin") || strstr(fname,".ogm") || strstr(fname,".mpg") || strstr(fname,".mpe"));
	} else {
		return (strstr(fname,".avi") || strstr(fname,".mpg") || strstr(fname,".ogm") || strstr(fname,".mpe"));
	}

	EMULE_CATCH

	return false;
}

bool CPartFile::IsAviMovie()
{
	EMULE_TRY

	// DonGato, updated by Vorlost
	CString fname = GetFileName();
	fname.MakeLower();

	return (strstr(fname,".avi") || strstr(fname,".ogm"));

	EMULE_CATCH

	return false;
}

bool CPartFile::IsMpgMovie()
{
	EMULE_TRY

	// DonGato, updated by Vorlost
	CString fname = GetFileName();
	fname.MakeLower();

	return (strstr(fname,".mpg") || strstr(fname,".mpeg"));

	EMULE_CATCH

	return false;
}

bool CPartFile::IsMovieChunk(uint8 mode, uint16 chunk)
{
	EMULE_TRY

	uint32 lastblock = ceil((float)PARTSIZE/BLOCKSIZE),end;
	if (chunk == GetPartCount()-2)
		end = (PARTSIZE*chunk) + (lastblock*BLOCKSIZE)-1; //Endadress from 1chunk before the last chunk

	switch (mode)
	{
 		case 0: //Mode 0: No Preview
			return false;
			break;
 		case 1: //Mode 1: 1F,1L(2L) AVI
  			if (chunk == 0)
				return true; //1F
  			if (chunk == GetPartCount()-1)
				return true; //1L is a moviechunk ?
  			if (chunk == (GetPartCount()-2) && (((GetFileSize()-1) - MINLASTMOVIESIZE ) <= end))
				return true; //2L is a Part of 1/8L at the end of the file
  			break;
 		case 2: //Mode 2: 1F,2F,1L(2L) Long AVI
  			if (chunk == 0)
				return true; //1F
  			if (chunk == 1)
				return true; //2F
  			if (chunk == GetPartCount()-1)
				return true; //1L is a moviechunk ?
  			if (chunk == (GetPartCount()-2) && (((GetFileSize()-1) - MINLASTMOVIESIZE ) <= end))
				return true; //2L is a Part of 1/8L at the end of the file
  				break;
 		case 3: //Mode 3: 1F,2F,3F,1L(2L) Longest AVI
  			if (chunk == 0)
				return true; //1F
  			if (chunk == 1)
				return true; //2F
  			if (chunk == 2)
				return true; //3F
  			if (chunk == GetPartCount()-1)
				return true; //1L is a moviechunk
  			if (chunk == (GetPartCount()-2) && (((GetFileSize()-1) - MINLASTMOVIESIZE ) <= end))
				return true; //2L is a Part of 1/8L at the end of the file
  				break;
 		case 4: //Mode 4: 1F MPG
  			if (chunk == 0)
				return true; //1F
  			break;
 		case 5: //Mode 5: 1F,2F Long MPG
  			if (chunk == 0)
				return true; //1F
  			if (chunk == 1)
				return true; //2F
  			break;
 		case 6: //Mode 6: 1F,2F,3F Longest MPG
  			if (chunk == 0)
				return true; //1F
  			if (chunk == 1)
				return true; //2F
  			if (chunk == 2)
				return true; //3F
  			break;
 		case 7: //Mode 7: 1F,2F,1L,2L AVI/MPG normal
  			if (chunk == 0)
				return true; //1F
  			if (chunk == 1)
				return true; //2F
  			if (chunk == GetPartCount()-1) 
				return true; //1L is a moviechunk ?
  			if (chunk == GetPartCount()-2)
				return true; //2L is a moviechunk ?
  			break;
 		default:
  			return false;
			break;
	}

	EMULE_CATCH

	return false;
}

bool CPartFile::IsBetterMovieChunk(uint16 oldchunk,uint16 newchunk)
{
	EMULE_TRY

	//The routine is based on that both old&new-chunk are moviechunks
	if (oldchunk == 0) //Chunk = 1F
	{
		// 1/8F & !1/8L & (newchunk=1L | 2L)
 		if (IsComplete(0,(MINFIRSTMOVIESIZE)-1) //first 1/8 chunk present
  			&& !IsComplete( (GetFileSize()-1) - (MINLASTMOVIESIZE) ,GetFileSize()-1) //last 1/8 chunk not present
  			&& (newchunk == GetPartCount()-1  //better chunk could be 1L
   			|| (newchunk == GetPartCount()-2))) //better chunk could be 2L
				return true;
	}
	else if (oldchunk == GetPartCount()-1 || oldchunk == GetPartCount()-2) //Chunk = 1L | 2L
	{
 		// (1/8L & !1/8F) | (1/8L & 1/8F)
 		if (((!IsComplete(0,(MINFIRSTMOVIESIZE)-1) //first 1/8 chunk not present
    		&& IsComplete( (GetFileSize()-1) - (MINLASTMOVIESIZE) ,GetFileSize()-1)) //last 1/8 chunk present
   			||(IsComplete(0,(MINFIRSTMOVIESIZE)-1) //first 1/8 chunk present
        	&& IsComplete( (GetFileSize()-1) - (MINLASTMOVIESIZE) ,GetFileSize()-1)) //last 1/8 chunk present
            && newchunk == 0))  //better chunk could be 1F
				return true;
	}
	else if (oldchunk == 1) //Chunk = 2F
	{
 		// 1F | 1L
 		if (newchunk == 0 || newchunk == GetPartCount()-1)
			return true;
	}
	else if (oldchunk == 2) //Chunk = 3F
	{
 		// 1F | 1L otherwise 2F
 		if (newchunk == 0 || newchunk == GetPartCount()-1)
			return true;
 		else if (newchunk == 1)
			return true;
	}

	EMULE_CATCH

	return false;
}

//Movie1 disables (movie --> normal) & (movie --> worst movie)
bool CPartFile::Movie1(uint8 mode,uint16 oldchunk,uint16 newchunk)
{
	EMULE_TRY

	if (IsMovie())
	{
 		if (!IsMovieChunk(mode,newchunk) && IsMovieChunk(mode,oldchunk))
  			return false;
 		if (IsMovieChunk(mode,newchunk) && IsMovieChunk(mode,oldchunk))
		{
  			if (!IsBetterMovieChunk(oldchunk,newchunk))
   				return false;
		}
	}
	return true;

	EMULE_CATCH

	return false;
}

//Movie2 enables (normal --> movie) & (movie --> better movie)
bool CPartFile::Movie2(uint8 mode,uint8 raremovie,bool forcepreview,uint16 oldchunk,uint16 newchunk)
{
	EMULE_TRY

	if (IsMovie() && ((m_SrcpartFrequency[oldchunk] > raremovie
    	&& m_SrcpartFrequency[newchunk] > raremovie)      //old & new chunk are nonrare
    	|| (m_SrcpartFrequency[oldchunk] == m_SrcpartFrequency[newchunk]) //old & new chunk are same rare
    	|| forcepreview))            //Load preview chunks first no matter of rareness
	{
			//This Line just for debugging:
			//theApp.emuledlg->AddLogLine(true,false,CString("Movie2::Mode:%d old:%d new:%d rarem:%d - The new & old ones are nonrare,samerare or forced, now we decide if new is a better movie or could change to movie..."),mode,oldchunk,newchunk,raremovie);
 			if (!IsMovieChunk(mode,oldchunk) && IsMovieChunk(mode,newchunk))
			{
  				if (forcepreview)
					return true;
 			}
 			if (IsMovieChunk(mode,oldchunk) && IsMovieChunk(mode,newchunk))
			{
  				if (IsBetterMovieChunk(oldchunk,newchunk))
				{
   					if (forcepreview)
	 					return true;
  				}
 			}
	}

	EMULE_CATCH

	return false;
}

uint8 CPartFile::PrepareComparePart(uint16 chunk)
{
	EMULE_TRY

	uint8 addchunk = 0;

	if (chunk == GetPartCount()-1) //Is this the last chunk ?
	{
 		uint32 lastblock = ceil((float)PARTSIZE/BLOCKSIZE);
 		uint32 lastblockinpart = lastblock;
 		uint32 start;
    	//Get last possible blocknumber
 		for (lastblockinpart = 0; lastblockinpart != lastblock;lastblockinpart++)
		{
  			start = (PARTSIZE*chunk) + lastblockinpart*BLOCKSIZE;
  			if (start >= GetFileSize())
   				break;
 		}
 		addchunk = lastblock - lastblockinpart;
	}

	return GetRemainingBlocksInPart(chunk) + addchunk;

	EMULE_CATCH

	return 0;
}
//Vorlost for LoadRarestFinishedChunks Patch2 end

// Barry - Also want to preview zip/rar files
bool CPartFile::IsArchive()
{
	CString extension = CString(GetFileName()).Right(4);
	return ((extension.CompareNoCase(".zip") == 0) || (extension.CompareNoCase(".rar") == 0));
}

void CPartFile::SetPriority(uint8 np)
{
	EMULE_TRY

	//UAP
	if (np == PR_AUTO)
		SetAutoPriority(true) ;
	else
		priority = np;
	//end UAP
	theApp.downloadqueue->SortByPriority();
	//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	UpdateDisplayedInfo(true);
	SavePartFile();

	EMULE_CATCH
}

void CPartFile::StopFile()
{
	EMULE_TRY

	// Barry - Need to tell any connected clients to stop sending the file
	PauseFile();

	theApp.emuledlg->transferwnd.downloadlistctrl.HideSources(this); // cax2 closefile bugfix
	RemoveAllSources(true);
	paused = true;
	datarate = 0;
	transferingsrc = 0;
	FlushBuffer();
	//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	UpdateDisplayedInfo(true);

	EMULE_CATCH
}

void CPartFile::PauseFile()
{
	EMULE_TRY

	if (status==PS_COMPLETE || status==PS_COMPLETING) return;
	Packet* packet = new Packet(OP_CANCELTRANSFER,0);
	POSITION pos1,pos2;
	for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
	for( pos1 = srclists[sl].GetHeadPosition(); ( pos2 = pos1 ) != NULL; ){
		srclists[sl].GetNext(pos1);
		CUpDownClient* cur_src = srclists[sl].GetAt(pos2);
		if (cur_src->GetDownloadState() == DS_DOWNLOADING){
			theApp.uploadqueue->AddUpDataOverheadOther(packet->size);
			cur_src->socket->SendPacket(packet,false,true);
			cur_src->SetDownloadState(DS_ONQUEUE);
		}
	}	
	delete packet;
	paused = true;
	datarate = 0;
	transferingsrc = 0;
	//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	UpdateDisplayedInfo(true);
	SavePartFile();

	EMULE_CATCH
}

void CPartFile::ResumeFile()
{
	EMULE_TRY

	if (status==PS_COMPLETE || status==PS_COMPLETING) return;
	paused = false;
	lastsearchtime = 0;
	//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	UpdateDisplayedInfo(true);
	
#ifdef OLD_SLS_ENABLED
	SetLoadSourcesTimeInterval(2 * 60 * 1000) ; // Added by Tarod (to load exchanged sources)
#endif //OLD_SLS_ENABLED

	EMULE_CATCH
}

// Added by Vorlost to load initial sources from Server after connect 
void CPartFile::GetSourcesAfterServerConnect()
{
	EMULE_TRY

	if (lastsearchtime>0){ 
		// theApp.emuledlg->AddLogLine(true,CString("Resetting LastSearchTime:%d in Filename:%s"),lastsearchtime,GetFileName());
		lastsearchtime = 0; 
	} 

	EMULE_CATCH
}

CString CPartFile::GetPartfileStatus()
{
	EMULE_TRY

	CString mybuffer=""; 
	if (GetTransferingSrcCount()>0) mybuffer=GetResString(IDS_DOWNLOADING); else mybuffer=GetResString(IDS_WAITING); 
	switch (GetStatus()) {
		case PS_HASHING: 
		case PS_WAITINGFORHASH:
			mybuffer=GetResString(IDS_HASHING);
			break; 
		case PS_COMPLETING:
			mybuffer=GetResString(IDS_COMPLETING);
			break; 
		case PS_COMPLETE:
			mybuffer=GetResString(IDS_COMPLETE);
			break; 
		case PS_PAUSED:
			mybuffer=GetResString(IDS_PAUSED);
			break; 
		case PS_ERROR:
			mybuffer=GetResString(IDS_ERRORLIKE);
			break;
	} 
	return mybuffer; 

	EMULE_CATCH

	return "";
} 

int CPartFile::getPartfileStatusRang()
{
	EMULE_TRY 

	int status=0;

	int tempstatus=0;
	if (GetTransferingSrcCount()==0)
		tempstatus=1;
	switch (GetStatus()) {
		case PS_HASHING: 
		case PS_WAITINGFORHASH:
			tempstatus=3;
			break; 
		case PS_COMPLETING:
			tempstatus=4;
			break; 
		case PS_COMPLETE:
			tempstatus=5;
			break; 
		case PS_PAUSED:
			tempstatus=2;
			break; 
		case PS_ERROR:
			tempstatus=6;
			break;
	} 
	return tempstatus;

	EMULE_CATCH

	return 1;
} 

sint32 CPartFile::GetTimeRemaining()
{
	EMULE_TRY

	if (GetDatarate()==0) return -1;
	
	return( (GetFileSize()-GetCompletedSize())/ GetDatarate());

	EMULE_CATCH

	return -1;
} 

void CPartFile::PreviewFile()
{
	EMULE_TRY

	if (IsArchive())
	{
		if ((!m_bRecoveringArchive) && (!m_bPreviewing))
			CArchiveRecovery::recover(this, true);
		return;
	}

	if (!PreviewAvailable())
	{
		ASSERT( false );
	 	return;
	}

	if(IsVLCInstalled())
	{
		CString strLine = this->GetFullName();
		strLine.Replace(".met","");
		char shortPath[512]; //Cax2 short path for vlc
		GetShortPathName(strLine,shortPath,512);
		ShellExecute(NULL, "open", "vlc.exe", shortPath, theApp.glob_prefs->GetVlcDir(), SW_SHOWNORMAL);
		//delete[] &shortPath; //Cax2 oops, another unneded delete...
		//delete strLine; 
	} else {
		m_bPreviewing = true;
		CPreviewThread* pThread = (CPreviewThread*) AfxBeginThread(RUNTIME_CLASS(CPreviewThread), THREAD_PRIORITY_NORMAL,0, CREATE_SUSPENDED);
		pThread->SetValues(this);
		pThread->ResumeThread();
	}

	EMULE_CATCH
}

bool CPartFile::PreviewAvailable()
{
	EMULE_TRY

	uint64 buf1,buf2,space;
	space = 0;
	GetDiskFreeSpaceEx(tempdir,(PULARGE_INTEGER)&space,(PULARGE_INTEGER)&buf1,(PULARGE_INTEGER)&buf2);
	// Barry - Allow preview of archives of any length > 1k
	if (IsArchive()) 
		if (GetStatus() != PS_COMPLETE &&  GetStatus() != PS_COMPLETING && ( GetFileSize() > 1024) && (!m_bRecoveringArchive) && ((space + 100000000) > (2*GetFileSize())))
			return true; else return false;

	//Vorlost for LoadRarestFinishedChunks Patch2
	bool first = false,last = false;

 	char buffer[200] = "";
 	if (IsMovie())
	{
  		//check if min. Movie-Blocks loaded
  		if (IsComplete(0,(MINFIRSTMOVIESIZE)-1))
			first = true; //first blocks present
  
		if (IsComplete((GetFileSize()-1) - (MINLASTMOVIESIZE) ,GetFileSize()-1 ))
			last = true; //last blocks present

  		if (first == false) //first present & last with avi
		{
   			strcat(buffer,", first ");
   			if (last == false && IsAviMovie())
				strcat(buffer,"& last ");
  		}
  		else if (last == false && IsAviMovie())
			strcat(buffer,", last ");

  		if (!first || (!last && IsAviMovie()))
			strcat(buffer,"Movieblocks not downloaded yet");
 	}

	if (!first || (!last && IsAviMovie()))
		return false; //Blocks missing to preview

	if (!IsMovie())
		return false;

	if (GetStatus() == PS_COMPLETING || GetStatus() == PS_COMPLETE)
		return false;

	if (GetStatus() == PS_HASHING)
		return false;

	if ((space + 100000000) < GetFileSize() && IsMovie() && !IsVLCInstalled())
		return false;

	if ( m_bPreviewing && IsMovie())
		return false;

	return true;
	//End of Patch by Vorlost 

	EMULE_CATCH

	return false;
}

void CPartFile::UpdateAvailablePartsCount(){
	EMULE_TRY

	uint8 availablecounter = 0;
	bool breakflag = false;
	uint16 iPartCount = GetPartCount();
	for (uint32 ixPart = 0; ixPart < iPartCount; ixPart++){
		breakflag = false;
		for (uint32 sl = 0; sl < SOURCESSLOTS && !breakflag; sl++){
			if (!srclists[sl].IsEmpty()){
				for(POSITION pos = srclists[sl].GetHeadPosition(); pos && !breakflag; ){
					if (srclists[sl].GetNext(pos)->IsPartAvailable(ixPart)){ 
						availablecounter++; 
						breakflag = true;
					}
				}
			}
		}
	}
	if (iPartCount == availablecounter && availablePartsCount < iPartCount)
		lastseencomplete = CTime::GetCurrentTime();
	availablePartsCount = availablecounter;

	EMULE_CATCH
}

Packet*	CPartFile::CreateSrcInfoPacket(CUpDownClient* forClient)
{
	EMULE_TRY

	int sl;
	for (sl=0;sl<SOURCESSLOTS;sl++) if (srclists[sl].IsEmpty()) return 0;

	CMemFile data;
	uint16 nCount = 0;

	data.Write(filehash,16);
	data.Write(&nCount,2);
	bool bNeeded;
	for (sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
	for (POSITION pos = srclists[sl].GetHeadPosition();pos != 0;srclists[sl].GetNext(pos)){
		bNeeded = false;
		CUpDownClient* cur_src = srclists[sl].GetAt(pos);
		// Modified by Tarod -sends sources with LowID that are in same sever
		// that client is requesting sources
		if (cur_src->HasLowID() 
			&& (cur_src->GetServerIP() != forClient->GetServerIP() || cur_src->GetServerPort() != forClient->GetServerPort()))
			continue;

		// Added by Tarod, do not send incomplete established conns. to avoid sending
		// no valid sources
		switch (cur_src->GetDownloadState()) {
			case DS_NONE:
			case DS_CONNECTING:
			case DS_ERROR:
			case DS_TOOMANYCONNS:
			continue;
				break;
		}

		// only send source which have needed parts for this client if possible
		if (forClient->reqfile == this && forClient->GetPartStatus() && cur_src->GetPartStatus()){
			uint8* reqstatus = forClient->GetPartStatus();
			uint8* srcstatus = cur_src->GetPartStatus();
			for (int x = 0; x != GetPartCount(); x++){
				if (srcstatus[x] && !reqstatus[x]){
					bNeeded = true;
					break;
				}
			}
		}
		if( bNeeded ){
                    nCount++;
                    uint32 dwID = cur_src->GetUserID();
                    uint16 nPort = cur_src->GetUserPort();
                    uint32 dwServerIP = cur_src->GetServerIP();
                    uint16 nServerPort = cur_src->GetServerPort();
                    data.Write(&dwID,4);
                    data.Write(&nPort,2);
                    data.Write(&dwServerIP,4);
                    data.Write(&nServerPort,2);
    		if (nCount > 500)
    			break;
            }
	}
	if (!nCount)
		return 0;
	data.Seek(16,0);
	data.Write(&nCount,2);

	Packet* result = new Packet(&data,OP_EMULEPROT);
	result->opcode = OP_ANSWERSOURCES;
	if (nCount > 28)
		result->PackPacket();
	theApp.emuledlg->AddDebugLogLine(false, "Send:Source User(%s) File(%s) Count(%i)", forClient->GetUserName(), GetFileName(), nCount);
	return result;

	EMULE_CATCH

	return NULL;
}

void CPartFile::AddClientSources(CMemFile* sources)
{
	EMULE_TRY

	//if this part file is paused or complete/completing no sources needed any more
	if(GetStatus()==PS_PAUSED || GetStatus() > 7)		//Cax2 + SyruS check
		return;

	uint16 nCount;
	sources->Read(&nCount,2);

	theApp.emuledlg->AddDebugLogLine(false, "RCV:Sources File(%s) Count(%i)", GetFileName(), nCount);

	if (theApp.glob_prefs->IsRemoveNNSEnabled())  //Vorlost added prefs question
	{
	    // Added by Tarod - calculate if we need more space than what is disponible
	    // if so remove No Needed Sources
	    int source_count_plus_new = GetSourceCount() + nCount;
	    if (source_count_plus_new > theApp.glob_prefs->GetMaxSourcePerFile()
		    && ((lastpurgetime == 0) || (::GetTickCount() - lastpurgetime) > PURGE_TIME)) {
    
		    RemoveNoNeededSources(source_count_plus_new - theApp.glob_prefs->GetMaxSourcePerFile());
	    }
	}

	for (int i = 0;i < nCount;i++){
		uint32 dwID;
		uint16 nPort;
		uint32 dwServerIP;
		uint16 nServerPort;
		sources->Read(&dwID,4);
		sources->Read(&nPort,2);
		sources->Read(&dwServerIP,4);
		sources->Read(&nServerPort,2);
		// check first if we are this source
		if (theApp.serverconnect->GetClientID() < 16777216 && theApp.serverconnect->IsConnected()){
			if ((theApp.serverconnect->GetClientID() == dwID) && theApp.serverconnect->GetCurrentServer()->GetIP() == dwServerIP)
				continue;
		}
		else if (theApp.serverconnect->GetClientID() == dwID)
			continue;
		// Modified by Tarod -use LowID source only if source is in our same server
		else if (dwID < 16777216 && !theApp.serverconnect->IsLocalServer(dwServerIP, nServerPort))
			continue;

		CUpDownClient* newsource = new CUpDownClient(nPort,dwID,dwServerIP,nServerPort,this);
		if(newsource == NULL)
			continue;
		if( theApp.glob_prefs->GetMaxSourcePerFile() > GetSourceCount()
			&& i < (int)theApp.glob_prefs->GetMaxSourcesPerFileSoftValue()){ // DonGato - Using Max Sources per File (soft)
				theApp.downloadqueue->CheckAndAddSource(this,newsource);
		} else {
			if(theApp.downloadqueue->IsSourcesLoadingEnabled())
			{
				CString strName;
				strName.Format(EXCHANGEDSOURCEMARK"%u", dwID);
#ifdef OLD_SLS_ENABLED
				if(m_stscrStoringSources != NULL)
					m_stscrStoringSources->UpdateSource(strName, nPort, dwID, dwServerIP, nServerPort, newsource->GetAvailableOfNeededPartsCount());
#endif //OLD_SLS_ENABLED
			}
			delete newsource;
		} 
	}

	EMULE_CATCH
}

void CPartFile::GetSizeToTransferAndNeededSpace(uint32& pui32SizeToTransfer, uint32& pui32NeededSpace) 
{
	EMULE_TRY

	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos)){
		Gap_Struct* cur_gap = gaplist.GetAt(pos);
		pui32SizeToTransfer += cur_gap->end - cur_gap->start;
		if(cur_gap->end == GetFileSize()-1)
			pui32NeededSpace = cur_gap->end - cur_gap->start;
	}

	EMULE_CATCH
}

///////////////////////////////////////////////////////////////////////////////////////
//Storing Sources:
#ifdef OLD_SLS_ENABLED
bool CPartFile::SaveSourcesToFile(CStoredSources* toSave, bool debug)
{
	EMULE_TRY

	//if this part-file is complete/completing no source saving is needed any more
	if(GetStatus() > 7)
		return false;

	//save the sources to disk
	if(toSave != NULL){
		CString temp="";
		temp.Format("%s\\%s.stsrc",tempdir,GetPartMetFileName());
		CFile saveSources;
		if(saveSources.Open(temp,CFile::modeCreate|
								//CFile::modeNoTruncate|
								CFile::shareDenyWrite|
								CFile::modeWrite,NULL)){
			CArchive archive(&saveSources, CArchive::store);
			archive.SetStoreParams(sizeof(CSourceEntry)*toSave->GetCountSourceEntrys()/10,sizeof(CStoredSources));
			try{
				toSave->Serialize(archive);
			}catch(CArchiveException* err){
				err->Delete();
				archive.Close();
				saveSources.Close();
				return false;
			}
			archive.Close();
			saveSources.Close();
			if(debug) theApp.emuledlg->AddLogLine(false,"Sources for '%s' (%s) were found and saved",GetPartMetFileName(),GetFileName());
			//TRACE("Saved sources for %s (%s.src)\n",storedSources->GetFileName(), storedSources->GetPartFileName());
		}
		else{
			TRACE("Could not open file for saving sources of %s (%s.stsrc)\n",toSave->GetFileName(), toSave->GetPartFileName());
			return false;
		}
	}
	else{
		if(debug) theApp.emuledlg->AddLogLine(false,"Due to the fact no sources were found for '%s', nothing was written",GetPartMetFileName());
		TRACE("No sources found for %s (%s)\n",toSave->GetFileName(), toSave->GetPartFileName());
	}

	TRACE("Saving sources for %s: all sources were saved once again since %.02f min. ago\n"
					,GetFileName()
					,(float(::GetTickCount() - m_iLastTimeSourcesSaved)/60000));

	m_iLastTimeSourcesSaved = ::GetTickCount();

	return true;

	EMULE_CATCH
	
	return false;
}

CStoredSources*	CPartFile::LoadSourcesFromFile(bool debug)
{
	EMULE_TRY

	CString temp="";
	temp.Format("%s\\%s.stsrc",tempdir,GetPartMetFileName());
	CFile loadSources;
	CStoredSources* toLoad = new CStoredSources(GetPartMetFileName()
												, GetFileName()
												, GetFileHash()
												, theApp.glob_prefs->GetMaxSourcePerFile());
	if(loadSources.Open(temp,CFile::modeRead|CFile::shareDenyWrite,NULL)){
		CArchive archive(&loadSources, CArchive::load);
		CFileStatus fileStatus;
		loadSources.GetStatus(fileStatus);
		long size = fileStatus.m_size;
		archive.SetLoadParams((2*size)/1000);
		try{
			toLoad->Serialize(archive);
		}catch(CException* err){
			char errorMsg[254];
			err->GetErrorMessage(errorMsg,254);
			if(debug) theApp.emuledlg->AddLogLine(false,"Could not load *.stsrc file for %s (%s)\r\n\t(ErrorMessage: %s)",GetFileName(),GetPartMetFileName(),errorMsg);
			err->Delete();
			archive.Close();
			loadSources.Close();
			delete toLoad;
			return NULL;
		}
		if(debug) theApp.emuledlg->AddLogLine(false,"Found %i Sources for %s in %s.stsrc",toLoad->GetCountSourceEntrys()
									,GetFileName(),toLoad->GetPartFileName());
		archive.Close();
		loadSources.Close();
	}
	else {
		if(debug) theApp.emuledlg->AddLogLine(false,"Could not open *.stsrc file for %s (%s)",GetFileName(),GetPartMetFileName());
		delete toLoad;
		return NULL;
	}

	//set the stored-sources preferences from file
	m_iLoadingSourcesTimeInterval			= toLoad->GetLoadSourcesTime();
	m_iLoadingSourcesPartialTimeInterval	= toLoad->GetLoadSourcesSlowTime();
	m_iLoadSourcesAtOnceLimit				= toLoad->GetLoadSourcesLimit();
	m_iSavingSourcesTimeInterval			= toLoad->GetSaveSourcesTime();
	m_bLoadSavedSources					= toLoad->GetLoadSources();
	
	return toLoad;

	EMULE_CATCH

	return NULL;
}

bool CPartFile::AddFromStoredSources(CStoredSources* storedSources, bool sendUDPRequest
									, int limit, bool reset, bool debug)
{
	EMULE_TRY

	//if this part-file is complete/completing no source-loading is needed any more
	if(GetStatus() > 7|| GetStatus()==PS_PAUSED)		//Cax2 + SyruS check
		return false;

	if (limit>0)limit = RESTORE_SOURCES_AT_ONCE; // Vorlost (bypass the old value stored in each SLS File) 

	if(storedSources != NULL){
		
		//if limit == -1 (unlimited), the index should not set to 0, because
		//the sources from index 0 to m_iSourceIndex where already
		//'slow-loaded' a short time ago (so no need to to it again)
		if(reset){
			storedSources->SetSourceIndex(0);
			m_iSourceIndex	= 0;
		}
		else 
			storedSources->SetSourceIndex(m_iSourceIndex);

		CSourceEntry* sourceEntry = storedSources->GetNextSourceEntry();//storedSources->GetFirstSourceEntry();
		CServerEntry* serverEntry;
		int countAddedSources=0;//count the added Sources
		int countUDP=0;			//count how many UDP-Requests were send (for debugging)
		uint32 userID;			//temporal variable for a User-ID

		//if sourceEntry is valid and no-limit (for limit equal -1)
		//or loop-counter i < limit, do...
		for(int i=0 ; (sourceEntry != NULL) && ((limit == -1)||(i < limit)); i++){
			serverEntry = sourceEntry->GetLastServerEntry();
			//TRACE("CDownloadQueue::LoadAndAddSourcesFromFile(...): UDP-Source-Request is %s!\n",(sendUDPRequest == true? "ON":"OFF"));
			
			////ask the known servers (of this client/source) for new sources
			//if(sendUDPRequest){
			//	//limited UDP-Requests to 5
			//	for(int i=0; (i < 5) && serverEntry != NULL ; i++){
			//	//while(serverEntry != NULL){
			//		if(theApp.serverconnect->IsConnected()){
			//			if (!theApp.listensocket->TooManySockets()){
			//				//is the server (for the request) already existing?
			//				CServer* server;
			//				char* serverIP = serverEntry->GetFullServerIP();
			//				server = theApp.serverlist->GetServerByAddress(
			//									serverIP
			//									,serverEntry->GetServerPort());
			//				delete serverIP;
			//				
			//				if(server != NULL){
			//					//if not, add the server
			//					//server = new CServer(serverEntry->GetServerPort(),serverEntry->GetFullServerIP());
			//					//theApp.emuledlg->serverwnd.serverlistctrl.AddServer(server);
			//				
			//					Packet packet(OP_GLOBGETSOURCES,16);
			//					memcpy(packet.pBuffer,GetFileHash(),16);
			//					theApp.serverconnect->SendUDPPacket(&packet,server,false);
			//					if(debug) theApp.emuledlg->AddLogLine(false,"sended an UDP-Source-Request to %s for %s (%s)\n"
			//														,server->GetFullIP()
			//														,GetFileName()
			//														,GetPartMetFileName());
			//					countUDP++;
			//				}//end: if(server != NULL)
			//			}//end: if(TooManySockets)
			//		}//end: if(isConnected)
			//		serverEntry = sourceEntry->GetPreviousServerEntry();
			//	}//end for-loop
			//}//end: if(sendUDPRequest)
			
			//generate CUpDownClients for known sources
			serverEntry = sourceEntry->GetLastServerEntry();			
			if(serverEntry !=  NULL){//you need a server for creating a CUpDownClient
				userID=sourceEntry->GetFirstUserID();
				if(m_iSourceIDIndex > 0 && limit != -1){
					for(int j=0; j < m_iSourceIDIndex; j++)
						userID = sourceEntry->GetNextUserID();
				}
				else m_iSourceIDIndex = 1;
				//TRACE("(%s)first source-index: %s:%i (ID: %u)\n",GetFileName()
				//	,sourceEntry->GetUserName(),m_iSourceIDIndex,userID);

				while(userID > 0){
					//check, if the sources-per-file-limit (glob_prefs) is exceeded
                    int srcsCount = 0;
                    for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
                        srcsCount += srclists[sl].GetCount();
					if(srcsCount >= theApp.glob_prefs->GetMaxSourcePerFile()
						||  ((limit != -1)&&(countAddedSources > limit)) ){
						goto After_Loading;
					}
					
					// Added by Tarod (LS tweak)
					// HACK by Md5Chap 2002-11-26 (Moved to CUpDownClient::~CUpDownClient)
					//sourceEntry->IncrementFailedConnections();
					CUpDownClient* newSource = new CUpDownClient(sourceEntry->GetUserPort()
														, userID, serverEntry->GetServerIP()
														, serverEntry->GetServerPort(), this, sourceEntry);
					
					CString strUserName = sourceEntry->GetUserName() ;
					if (strUserName.Find(GetResString(IDS_EXCHANGEDSOURCEMARK)) == 0) {
						strUserName = GetResString(IDS_EXCHANGEDSOURCE) ;
					} else {
						strUserName = GetResString(IDS_LOADEDSOURCE) ;
					}
					char* pName = new char [strUserName.GetLength() + 2] ;
					memset(pName, '\0', (strUserName.GetLength() + 2)) ;
					strncpy(pName, strUserName.GetBuffer(), strUserName.GetLength()) ;
					newSource->SetUserName(pName);
					delete [] pName; // FoRcHa

					theApp.downloadqueue->CheckAndAddSource(this, newSource, false);
					//TRACE("(LoadNAdd) new source (count: %i, index: %i) for %s's sended to CheckNAdd\n"
					//		//, (newSource->GetUserName()!=NULL?newSource->GetUserName():NONAME)
					//		, countAddedSources
					//		, m_iSourceIDIndex++
					//		, GetFileName());
					countAddedSources++;
					m_iSourceIDIndex++;
					if(debug){ theApp.emuledlg->AddLogLine(false,"added Source (%i) of %i Sources for %s (%s)"
							,newSource->GetUserID()
							,storedSources->GetCountSourceEntrys(),GetFileName()
							,storedSources->GetFileID());
						theApp.emuledlg->AddLogLine(false,"The new Source is %s and the SourceEntry is %s",(newSource == NULL ? "NULL" : "not NULL"),(sourceEntry == NULL ? "NULL" : "not NULL"));
						theApp.emuledlg->AddLogLine(false,"New source No. %i added for %s",countAddedSources,GetFileName());
					}
					 userID = sourceEntry->GetNextUserID();
				}//end while-loop
			}//end if(serverEntry != NULL)
			sourceEntry = storedSources->GetNextSourceEntry();//limit);
			m_iSourceIDIndex = -1;
//#ifdef _DEBUG
//            if(sourceEntry != NULL) TRACE("(%s)last source-index: %s:%i (ID: %u)\n",GetFileName()
//					,sourceEntry->GetUserName(),m_iSourceIDIndex,userID);
//#endif
		}//end while-loop
#ifdef _DEBUG
		if(sendUDPRequest && countUDP > 0)TRACE("###############################################\n%i UDP-Source-Requests for %s were sended\n###############################################\n",countUDP,GetFileName());
#endif
	}//end if(storedSources != NULL)
After_Loading:
	//for 'slow-loading' the sources
	bool isCompletelyLoaded = true;
	if(storedSources != NULL){

		m_iSourceIndex = storedSources->GetSourceIndex();

		//are all sources from the CStoredSources-object loaded?
		if(m_iSourceIndex +1 >= storedSources->GetCountSourceEntrys())
			isCompletelyLoaded = true;
		else
			isCompletelyLoaded = false;

		m_bLoadedSourcesCompletely = isCompletelyLoaded;

		//TRACE("Added the next %i sources up to source-index %u for %s\n"
		//		, (limit != -1?limit: (-1))
		//		, m_iSourceIndex
		//		, GetFileName());
	}

	//set timers
	if(limit != -1){
#ifdef _DEBUG
		if(storedSources != NULL) TRACE("Slow-loading %s (No. %u of %u): %u secs since last call\n"
				, GetFileName()
				, m_iSourceIndex
				, storedSources->GetCountSourceEntrys()
				,(::GetTickCount() - m_iLastTimeSourcesLoadedPartial)/1000);
#endif
		
		m_iLastTimeSourcesLoadedPartial = ::GetTickCount();
		if(isCompletelyLoaded){
#ifdef _DEBUG
			if(storedSources != NULL) TRACE("Slow-loading %s: all sources (%u of %u) were loaded once again since %.02f min. ago\n"
					, GetFileName()
					, m_iSourceIndex
					, storedSources->GetCountSourceEntrys()
					, (float(::GetTickCount() - m_iLastTimeSourcesLoaded)/60000));
#endif
		
			m_iLastTimeSourcesLoaded = ::GetTickCount();
		}
	}
	else{
		TRACE("Normal-loading %s: all sources (%u) were loaded once again since %.02f min. ago\n"
					, GetFileName()
					, m_iSourceIndex
					, (float(::GetTickCount() - m_iLastTimeSourcesLoaded)/60000));

		m_iLastTimeSourcesLoaded = ::GetTickCount();
	}
	
	//TRACE("End of trying to add sources from StoredSource (AddFromStoredSources)\n");
	if (isCompletelyLoaded) m_iLoadingSourcesTimeInterval = RESTORE_SOURCES_TIME ; // Added by Tarod

	return isCompletelyLoaded;

	EMULE_CATCH

	return false;
}

bool CPartFile::LoadAndAddSources(bool keepSourcesOpen, bool sendUDPRequest
									, int limit, bool reset, bool debug,bool addImmediately)
{
	EMULE_TRY

	//if this part file is paused or complete/completing no sources needed any more
	if(GetStatus()==PS_PAUSED || GetStatus() > 7)		//Cax2 + SyruS check
		return false;

	CStoredSources* storedSources = NULL;
	if((storedSources = LoadSourcesFromFile(debug)) == NULL){
		if(m_stscrStoringSources == NULL){
				m_stscrStoringSources = new CStoredSources(GetPartMetFileName()
													, GetFileName()
													, GetFileHash()
													, theApp.glob_prefs->GetMaxSourcePerFileSoft());
				//set the preferences for the StoredSources
				m_stscrStoringSources->SetPreferences(theApp.glob_prefs->GetMaxSourcePerFileSoft()
												, m_iLoadSourcesAtOnceLimit
												, m_iLoadingSourcesTimeInterval
												, m_iLoadingSourcesPartialTimeInterval
												, m_iSavingSourcesTimeInterval);
		}
		return false;
	}
	//if successfully loaded:
	else{
		if(addImmediately)AddFromStoredSources(storedSources, sendUDPRequest, limit, reset, debug);

		TRACE("There were %u sources loaded from disk for %s are %s\n"
				,storedSources->GetCountOfSources()
				,GetFileName()
				,((keepSourcesOpen&& theApp.downloadqueue->KeepStoredSourcesOpen())?"kept open!":"not kept open!"));
		if(keepSourcesOpen && theApp.downloadqueue->KeepStoredSourcesOpen()){
			if(m_stscrStoringSources == NULL){
				m_stscrStoringSources = new CStoredSources(GetPartMetFileName()
													, GetFileName()
													, GetFileHash()
													, theApp.glob_prefs->GetMaxSourcePerFileSoft());
				//set the preferences for the StoredSources
				m_stscrStoringSources->SetPreferences(theApp.glob_prefs->GetMaxSourcePerFileSoft()
												, m_iLoadSourcesAtOnceLimit
												, m_iLoadingSourcesTimeInterval
												, m_iLoadingSourcesPartialTimeInterval
												, m_iSavingSourcesTimeInterval);
			}
			m_stscrStoringSources->UpdateSource(storedSources);
		}
		delete storedSources;
	}
    return true;

	EMULE_CATCH

	return false;
}

bool CPartFile::SaveToStoredSources(CUpDownClient* toSave)
{
	EMULE_TRY

	if(!theApp.downloadqueue->IsSourcesLoadingEnabled())
		return false;

	if(m_stscrStoringSources == NULL){
		m_stscrStoringSources = new CStoredSources(GetPartMetFileName()
											, GetFileName()
											, GetFileHash()
											, theApp.glob_prefs->GetMaxSourcePerFileSoft());
		//set the preferences for the StoredSources
		m_stscrStoringSources->SetPreferences(theApp.glob_prefs->GetMaxSourcePerFileSoft()
										, m_iLoadSourcesAtOnceLimit
										, m_iLoadingSourcesTimeInterval
										, m_iLoadingSourcesPartialTimeInterval
										, m_iSavingSourcesTimeInterval);
	}
#ifdef _DEBUG_SAVED_SOURCES
	uint32 oldCount = m_stscrStoringSources->GetCountOfSources();
#endif
	uint16 partsCount = 0;
	CString userName;
	partsCount	= toSave->GetAvailableOfNeededPartsCount();	//get the count of parts, this user has and we want to download
	// userName	= (toSave->GetUserName() != NULL ? toSave->GetUserName(): NONAME);//check, if the user-name of this user is set/known
	userName.Format(LOADEDSOURCEMARK"%u", toSave->GetUserID());
	m_stscrStoringSources->UpdateSource(userName
													, toSave->GetUserPort()
													, toSave->GetUserID()
													, toSave->GetServerIP()
													, toSave->GetServerPort()
													, partsCount);
#ifdef _DEBUG_SAVED_SOURCES
	TRACE("CPartFile::SaveToStoredSources(...) : source was %s to %s's StoredSources (count (old/new): %u/%u\n\n"
		, (m_stscrStoringSources->FindSource(userName,toSave->GetServerPort())?"saved":"not saved")
		, GetFileName()
		, oldCount
		, m_stscrStoringSources->GetCountOfSources());
#endif
	return true;

	EMULE_CATCH

	return false;
}

bool CPartFile::SaveToStoredSources()
{
	EMULE_TRY

	if(!theApp.downloadqueue->KeepStoredSourcesOpen())
		return false;

	bool successful=false;
	if(m_stscrStoringSources == NULL){
		m_stscrStoringSources = new CStoredSources(GetPartMetFileName()
										, GetFileName()
										, GetFileHash()
										, theApp.glob_prefs->GetMaxSourcePerFileSoft());
		//set the preferences for the StoredSources
		m_stscrStoringSources->SetPreferences(theApp.glob_prefs->GetMaxSourcePerFileSoft()
											, m_iLoadSourcesAtOnceLimit
											, m_iLoadingSourcesTimeInterval
											, m_iLoadingSourcesPartialTimeInterval
											, m_iSavingSourcesTimeInterval);
	}
#ifdef _DEBUG_SAVED_SOURCES
	uint32 oldCount = m_stscrStoringSources->GetCountOfSources();
#endif
	uint16 partsCount = 0;
	CString userName;
	CUpDownClient* source;
	successful = false;	//just to be sure
    for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
    for(POSITION posOfSource=srclists[sl].GetHeadPosition() ; posOfSource != NULL;){
		source = srclists[sl].GetNext(posOfSource);
		partsCount = source->GetAvailableOfNeededPartsCount();	//get the count of parts, this user has and we want to download
		// userName = (source->GetUserName() != NULL ? source->GetUserName(): NONAME);//check, if the user-name of this user is set/known
		userName.Format(LOADEDSOURCEMARK"%u", source->GetUserID());
		m_stscrStoringSources->UpdateSource(userName, source->GetUserPort(), 
									source->GetUserID(), source->GetServerIP(), 
									source->GetServerPort(), partsCount);
		successful=true;
	}
#ifdef _DEBUG_SAVED_SOURCES
	TRACE("CPartFile::SaveToStoredSources() : all current sources of %s were %s to its StoredSources (count (old/new): %u/%u\n\n"
		, GetFileName()
		, successful?"saved":"not saved"
		, oldCount
		, m_stscrStoringSources->GetCountOfSources());
#endif
	return successful;

	EMULE_CATCH

	return false;
}

bool CPartFile::SaveSources(bool loadSourcesFirst)
{
	EMULE_TRY

	if(m_stscrStoringSources == NULL)
		return false;

	bool successful=false;
	if(theApp.downloadqueue->KeepStoredSourcesOpen() && theApp.downloadqueue->IsSourcesLoadingEnabled())
	{
		if(!SaveToStoredSources() && m_stscrStoringSources->GetCountOfSources() < 1) 
			return false;

		TRACE("%u sources for %s are about to be saved to disk\n"
				,m_stscrStoringSources->GetCountOfSources()
				,GetFileName());

		//no need to load the file again, because all its informations
		//are already in m_stscrStoringSources
		//no need to gather all sources from scrlist, because everytime
		//a source is added there, it is also added to m_stscrStoringSources
		//(see CDownloadQueue::CheckAndAdd[Known]Source(...))
		SaveSourcesToFile(m_stscrStoringSources);
		successful = true;
	}
	else if(theApp.downloadqueue->IsSourcesLoadingEnabled()){
		CStoredSources* storedSources = NULL;

		if(loadSourcesFirst){
			storedSources = LoadSourcesFromFile();
		}
		if(storedSources == NULL){
			storedSources = new CStoredSources(GetPartMetFileName()
											, GetFileName()
											, GetFileHash()
											, theApp.glob_prefs->GetMaxSourcePerFileSoft());
			//set the preferences for the StoredSources
			storedSources->SetPreferences(theApp.glob_prefs->GetMaxSourcePerFileSoft()
												, m_iLoadSourcesAtOnceLimit
												, m_iLoadingSourcesTimeInterval
												, m_iLoadingSourcesPartialTimeInterval
												, m_iSavingSourcesTimeInterval);
		}
		uint16 partsCount = 0;
		CString userName;
		CUpDownClient* source;
		successful = false;	//just to be sure
        for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
		for(POSITION posOfSource=srclists[sl].GetHeadPosition() ; posOfSource != NULL;){
			source = srclists[sl].GetNext(posOfSource);
			partsCount = source->GetAvailableOfNeededPartsCount();	//get the count of parts, this user has and we want to download
			// userName = (source->GetUserName() != NULL ? source->GetUserName(): NONAME);//check, if the user-name of this user is set/known
			userName.Format(LOADEDSOURCEMARK"%u", source->GetUserID());
			storedSources->UpdateSource(userName, source->GetUserPort(), 
										source->GetUserID(), source->GetServerIP(), 
										source->GetServerPort(), partsCount);
			successful=true;
		}
		
		//if(debug) theApp.emuledlg->AddLogLine(false,"All Sources of '%s' (%s) were gathered for saving to disk",GetFileName(),GetPartMetFileName());

		if(successful){
			SaveSourcesToFile(storedSources);
			TRACE("%u sources for %s were saved to disk with no permanent open StoredSources-object\n"
					,storedSources->GetCountSourceEntrys()
					,GetFileName());
		}
		delete storedSources;
	}
	else if(!theApp.downloadqueue->IsSourcesLoadingEnabled()){
		CStoredSources* storedSources = LoadSourcesFromFile();

		if(storedSources == NULL){
			storedSources = new CStoredSources(GetPartMetFileName()
											, GetFileName()
											, GetFileHash()
											, theApp.glob_prefs->GetMaxSourcePerFileSoft());
			//set the preferences for the StoredSources
			storedSources->SetPreferences(theApp.glob_prefs->GetMaxSourcePerFileSoft()
												, m_iLoadSourcesAtOnceLimit
												, m_iLoadingSourcesTimeInterval
												, m_iLoadingSourcesPartialTimeInterval
												, m_iSavingSourcesTimeInterval);
		}
		uint16 partsCount = 0;
		CString userName;
		CUpDownClient* source;
		successful = false;	//just to be sure
        for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
		for(POSITION posOfSource=srclists[sl].GetHeadPosition() ; posOfSource != NULL;){
			source = srclists[sl].GetNext(posOfSource);
			partsCount = source->GetAvailableOfNeededPartsCount();	//get the count of parts, this user has and we want to download
			// userName = (source->GetUserName() != NULL ? source->GetUserName(): NONAME);//check, if the user-name of this user is set/known
			userName.Format(LOADEDSOURCEMARK"%u", source->GetUserID());
			storedSources->UpdateSource(userName, source->GetUserPort(), 
										source->GetUserID(), source->GetServerIP(), 
										source->GetServerPort(), partsCount);
			successful=true;
		}
		
		//if(debug) theApp.emuledlg->AddLogLine(false,"All Sources of '%s' (%s) were gathered for saving to disk",GetFileName(),GetPartMetFileName());

		if(successful){
			SaveSourcesToFile(storedSources);
			TRACE("%u sources for %s were saved to disk while loading of sources is disabled\n"
					,storedSources->GetCountSourceEntrys()
					,GetFileName());
		}
		delete storedSources;
	}
	return successful;

	EMULE_CATCH

	return false;
}
#endif //OLD_SLS_ENABLED
// Sets this file as an A4AF auto
//
// [In]
//   in_A4AF_auto: true -> Set this file as A4AF auto
//                 false-> Unset this file as A4AF auto
//
// [Global]
//   in_A4AF_auto: true -> File hash is copied into prefs->A4AF_file_hash
//                 false-> "Null" hash is copied into prefs->A4AF_file_hash
void CPartFile::SetA4AFAuto(bool in_A4AF_auto)
{
	EMULE_TRY

	m_is_A4AF_auto = in_A4AF_auto;

	// Set file hash
	if (in_A4AF_auto) {
		theApp.glob_prefs->SetA4AFHash(GetFileHash());
	} else {
		// Set null hash
		uchar null_hash[16];
		memset(null_hash, 0, sizeof(null_hash));
		theApp.glob_prefs->SetA4AFHash(null_hash);
	}

	EMULE_CATCH
}

// Updates priority of file if autopriority is activated
void CPartFile::UpdateDownloadAutoPriority(void)
{
	EMULE_TRY

	if (IsAutoPrioritized()) {
		CPartFile* partfile = theApp.downloadqueue->GetFileByID(GetFileHash()) ;
		
		if (partfile != NULL && partfile->IsPartFile()) {
			int valid_sources = partfile->GetSourceCount() - partfile->GetNotCurrentSourcesCount();
			if (valid_sources < theApp.glob_prefs->PriorityHigh() && priority != PR_HIGH)
				SetPriority(PR_HIGH) ;
			else if (valid_sources >= theApp.glob_prefs->PriorityHigh() && valid_sources < theApp.glob_prefs->PriorityLow() && priority != PR_NORMAL)
				SetPriority(PR_NORMAL) ;
			else if (valid_sources >= theApp.glob_prefs->PriorityLow() && priority != PR_LOW)
				SetPriority(PR_LOW) ;
		} 
	}

	EMULE_CATCH
}

// making this function return a higher when more sources have the extended
// protocol will force you to ask a larger variety of people for sources
int CPartFile::GetCommonFilePenalty() {
	//TODO: implement, but never return less than MINCOMMONPENALTY!
	return MINCOMMONPENALTY;
}

//--- xrmb:keepPartFileStats ---
bool CPartFile::LoadPartFileStats()
{
	char	fn[MAX_PATH];
	strcpy(fn, fullname);
	fn[strlen(fn)-4]=0;	// 4 = .met
	strcat(fn, ".stats");
	FILE	*fh=fopen(fn, "rb");
	
	//--- if we cant open this file... you lose the stats ---
	if(fh==NULL)
		return false;
	
	//--- get version ---
	uint8	version=0;
	if(fread(&version, sizeof(version), 1, fh)!=1) { fclose(fh); return false; }

	uint32 alltimerequested=0;
	uint32 alltimeaccepted=0;
	uint64 alltimetransfered=0;

	if(fread(&alltimerequested,		sizeof(alltimerequested),	1, fh)!=1) { fclose(fh); return false; }
	if(fread(&alltimeaccepted,		sizeof(alltimeaccepted),	1, fh)!=1) { fclose(fh); return false; }
	if(fread(&alltimetransfered,	sizeof(alltimetransfered),	1, fh)!=1) { fclose(fh); return false; }

	statistic.AddAll(alltimerequested, alltimeaccepted, alltimetransfered);

	//--- this is for the parttraffic ---
	//--- i thought about explicit adding the traffic like for the other stats above ---
	//--- but it doesnt make sense, because you'll no transfer if you reach this ---
	if(LoadFromFileTraffic(fh, version)==false) { fclose(fh); return false;	}

	fclose(fh);

	return true;
}

bool CPartFile::SavePartFileStats()
{
	char	fn[MAX_PATH];
	strcpy(fn, fullname);
	fn[strlen(fn)-4]=0;	// 4 = .met
	strcat(fn, ".stats");
	FILE	*fh=fopen(fn, "wb");
	
	//--- if we cant open this file... you lose the stats ---
	if(fh==NULL)
		return false;
	
	//--- set version ---
	uint8	version=2;
	if(fwrite(&version, sizeof(version), 1, fh)!=1) { fclose(fh); return false; }

	uint32 alltimerequested=statistic.GetAllTimeRequests();
	uint32 alltimeaccepted=statistic.GetAllTimeAccepts();
	uint64 alltimetransfered=statistic.GetAllTimeTransfered();

	if(fwrite(&alltimerequested,	sizeof(alltimerequested),	1, fh)!=1) { fclose(fh); return false; }
	if(fwrite(&alltimeaccepted,		sizeof(alltimeaccepted),	1, fh)!=1) { fclose(fh); return false; }
	if(fwrite(&alltimetransfered,	sizeof(alltimetransfered),	1, fh)!=1) { fclose(fh); return false; }

	//--- this is for the parttraffic ---
	if(SaveToFileTraffic(fh)==false) { fclose(fh); return false; }

	fclose(fh);

	return true;
}

int CPartFile::RemovePartFileStats()
{
	char	fn[MAX_PATH];
	strcpy(fn, fullname);
	fn[strlen(fn)-4]=0;	// 4 = .met
	strcat(fn, ".stats");
	
	return remove(fn);
}
//--- :xrmb ---

// Used as separate thread to complete file [Lord KiRon]
UINT CPartFile::CompleteThreadProc(CPartFile* pFile) 
{ 
	EMULE_TRY

	SetThreadLocale(theApp.glob_prefs->GetLanguageID());

	if ( !pFile ) 
   	{ 
    	return -1; 
   	} 
	CSingleLock Lock(&s_ComplitingMutex,TRUE); // Prevent of compliting two files in same time
   	pFile->PerformFileComplete(); 
   	return 0; 

	EMULE_CATCH

   	return -1; 
}

bool CPartFile::IsVLCInstalled() 
{ 
	EMULE_TRY

	CFileFind ff;
	CString searchpath(theApp.glob_prefs->GetVlcDir());
	searchpath += "\\*";

	bool end = !ff.FindFile(searchpath,0);
	bool vlcfound = false;

	if (end)
	 	return false;
	while (!end)
	{
	 	end = !ff.FindNextFile();
		if (ff.IsDirectory())
   			continue;
 	 	if (ff.GetFileName().Find("vlc.exe") != -1)
 	 	{
			ff.Close();
			vlcfound = true;
		} else
			continue;
	}
	ff.Close();

	if(vlcfound)
		return true;
	else
		return false;

	EMULE_CATCH

	return false;
}	

// DonGato - Moview Preview Mode
void CPartFile::LoadMovieMode()
{ 
	EMULE_TRY

   	CIni pvmoviefile(CString(fullname) + CString(".movie"), "MoviePreview"); 
   	uint8 Mode = pvmoviefile.GetInt("Mode", 0); 
	CKnownFile::SetMovieMode(Mode);

	EMULE_CATCH
}

void CPartFile::SaveMovieMode()
{
	EMULE_TRY

   	CIni pvmoviefile(CString(fullname) + CString(".movie"), "MoviePreview"); 
	pvmoviefile.WriteInt("Mode", ((CKnownFile*)this)->GetMovieMode()); 

	EMULE_CATCH
}
// DonGato - Moview Preview Mode end

/* Barry - Replaces BlockReceived() 

           Originally this only wrote to disk when a full 180k block 
           had been received from a client, and only asked for data in 
		   180k blocks.

		   This meant that on average 90k was lost for every connection
		   to a client data source. That is a lot of wasted data.

		   To reduce the lost data, packets are now written to a buffer
		   and flushed to disk regularly regardless of size downloaded.
		   This includes compressed packets.

		   Data is also requested only where gaps are, not in 180k blocks.
		   The requests will still not exceed 180k, but may be smaller to
		   fill a gap.
*/
uint32 CPartFile::WriteToBuffer(uint32 transize, BYTE *data, uint32 start, uint32 end, Requested_Block_Struct *block)
{
	EMULE_TRY

	// Increment transfered bytes counter for this file
	transfered += transize;

	// This is needed a few times
	uint32 lenData = end - start + 1;

	if( lenData > transize )
		m_iGainDueToCompression += lenData-transize;

	// Occasionally packets are duplicated, no point writing it twice
	if (IsComplete(start, end))
	{
		theApp.emuledlg->AddDebugLogLine(false, "File '%s' has already been written from %ld to %ld\n", GetFileName(), start, end);
		return 0;
	}

	// Create copy of data as new buffer
	BYTE *buffer = new BYTE[lenData];
	memcpy(buffer, data, lenData);

	// Create a new buffered queue entry
	PartFileBufferedData *item = new PartFileBufferedData;
	item->data = buffer;
	item->start = start;
	item->end = end;
	item->block = block;

	// Add to the queue in the correct position (most likely the end)
	PartFileBufferedData *queueItem;
	bool added = false;
	POSITION pos = m_BufferedData_list.GetTailPosition();
	while (pos != NULL)
	{	
		queueItem = m_BufferedData_list.GetPrev(pos);
		if (item->end > queueItem->end)
		{
			added = true;
			m_BufferedData_list.InsertAfter(pos, item);
			break;
		}
	}
	if (!added)
		m_BufferedData_list.AddHead(item);

	// Increment buffer size marker
	m_nTotalBufferData += lenData;

	// Mark this small section of the file as filled
	FillGap(item->start, item->end);

	// Update the flushed mark on the requested block 
	// The loop here is unfortunate but necessary to detect deleted blocks.
	pos = requestedblocks_list.GetHeadPosition();
	while (pos != NULL)
	{	
		if (requestedblocks_list.GetNext(pos) == item->block)
			item->block->transferred += lenData;
	}

	if (gaplist.IsEmpty()) FlushBuffer();

	// Return the length of data written to the buffer
	return lenData;

	EMULE_CATCH

	return 0;
}

void CPartFile::FlushBuffer(void)
{
	EMULE_TRY

	m_nLastBufferFlushTime = GetTickCount();  // obaldin
	if (m_BufferedData_list.IsEmpty())
		return;

    m_lastDownTransfer = CTime::GetCurrentTime();
	//int partCount = filesize / PARTSIZE;
	// there are a couple of ways of doing this, this is one... 
	uint32 partCount = (filesize/PARTSIZE)+((filesize%PARTSIZE)? 1 : 0);
	
	bool *changedPart = new bool[partCount];

//	theApp.emuledlg->AddDebugLogLine(false, "Flushing file %s - buffer size = %ld bytes (%ld queued items) transfered = %ld [time = %ld]\n", GetFileName(), m_nTotalBufferData, m_BufferedData_list.GetCount(), transfered, m_nLastBufferFlushTime);

	try
	{
		// Remember which parts need to be checked at the end of the flush
		for (int partNumber=0; (uint32)partNumber<partCount; partNumber++)
		{
			changedPart[partNumber] = false;
		}

		// Ensure file is big enough to write data to (the last item will be the furthest from the start)
		PartFileBufferedData *item = m_BufferedData_list.GetTail();
		if (m_hpartfile.GetLength() <= item->end)
			m_hpartfile.SetLength(item->end + 1);

		// Loop through queue
		for (int i = m_BufferedData_list.GetCount(); i>0; i--)
		{
			// Get top item
			item = m_BufferedData_list.GetHead();

			// This is needed a few times
			uint32 lenData = item->end - item->start + 1;

			int curpart = item->start/PARTSIZE;
			changedPart[curpart] = true;

			// Go to the correct position in file and write block of data
			m_hpartfile.Seek(item->start, CFile::begin);
			m_hpartfile.Write(item->data, lenData);

			// Remove item from queue
			m_BufferedData_list.RemoveHead();

			// Decrease buffer size
			m_nTotalBufferData -= lenData;

			// Release memory used by this item
			delete [] item->data;
			delete item;
		}

		// Flush to disk
		m_hpartfile.Flush();

		// Check each part of the file
		uint32 partRange = (m_hpartfile.GetLength() % PARTSIZE) - 1;
		for (int partNumber = partCount-1; partNumber >= 0; partNumber--)
		{
			if (changedPart[partNumber] == false)
			{
				// Any parts other than last must be full size
				partRange = PARTSIZE - 1;
				continue;
			}

			// Is this 9MB part complete
			if ( IsComplete(PARTSIZE * partNumber, (PARTSIZE * (partNumber + 1)) - 1 ) )
			{
				// Is part corrupt
				if (!HashSinglePart(partNumber))
				{
					theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_PARTCORRUPT), partNumber, GetFileName());
					AddGap(PARTSIZE*partNumber, (PARTSIZE*partNumber + partRange));
					corrupted_list.AddTail(partNumber);
					// Reduce transfered amount by corrupt amount
					this->m_iLostDueToCorruption += (partRange + 1);
				}
				else
				{
					// Successfully completed part, make it available for sharing
					if (status == PS_EMPTY)
					{
						status = PS_READY;
						theApp.sharedfiles->SafeAddKFile(this);
					}
				}
			}
			else if ( IsCorruptedPart(partNumber))
			{
				// Try to recover with minimal loss
				if (HashSinglePart(partNumber))
				{
					m_iTotalPacketsSavedDueToICH++;
					FillGap(PARTSIZE*partNumber,(PARTSIZE*partNumber+partRange));
					RemoveBlockFromList(PARTSIZE*partNumber,(PARTSIZE*partNumber + partRange));
					theApp.emuledlg->AddLogLine(true,GetResString(IDS_ICHWORKED),partNumber,GetFileName());
				}
			}
			// Any parts other than last must be full size
			partRange = PARTSIZE - 1;
		}

		// Update met file
		SavePartFile();

		// Is this file finished?
		if (gaplist.IsEmpty())
			CompleteFile(false);
	}
	catch (CFileException* error)
	{
		OUTPUT_DEBUG_TRACE();

		if (theApp.glob_prefs->IsErrorBeepEnabled()) Beep(800,200);

		if (error->m_cause == CFileException::diskFull) 
		{
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_OUTOFSPACE), GetFileName());
			if (theApp.glob_prefs->GetNotifierPopOnImportantError()) 
			{
				CString msg;
				msg.Format(GetResString(IDS_ERR_OUTOFSPACE), GetFileName());
				theApp.emuledlg->ShowNotifier(msg, TBN_IMPORTANTEVENT, false);
			}
		}
		else
		{
			char buffer[150];
			error->GetErrorMessage(buffer,150);
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_WRITEERROR), GetFileName(), buffer);
			status = PS_ERROR;
		}
        /*
		paused = true;
		datarate = 0;
		transferingsrc = 0;
	
		//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
		UpdateDisplayedInfo();
        */
        PauseFile();
		error->Delete();
	}
	catch(...)
	{
		theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_WRITEERROR), GetFileName(), GetResString(IDS_UNKNOWN));
		status = PS_ERROR;
        PauseFile();
	}
	delete[] changedPart;

	EMULE_CATCH
}

// Barry - This will invert the gap list, up to caller to delete gaps when done
// 'Gaps' returned are really the filled areas, and guaranteed to be in order
void CPartFile::GetFilledList(CTypedPtrList<CPtrList, Gap_Struct*> *filled)
{
	EMULE_TRY

	Gap_Struct *gap;
	Gap_Struct *best;
	POSITION pos;
	uint32 start = 0;
	uint32 bestEnd = 0;

	// Loop until done
	bool finished = false;
	while (!finished)
	{
		finished = true;
		// Find first gap after current start pos
		bestEnd = filesize;
		pos = gaplist.GetHeadPosition();
		while (pos != NULL)
		{
			gap = gaplist.GetNext(pos);
			if ((gap->start > start) && (gap->end < bestEnd))
			{
				best = gap;
				bestEnd = best->end;
				finished = false;
			}
		}

		if (!finished)
		{
			// Invert this gap
			gap = new Gap_Struct;
			gap->start = start;
			gap->end = best->start - 1;
			start = best->end + 1;
			filled->AddTail(gap);
		}
		else if (best->end < filesize)
		{
			gap = new Gap_Struct;
			gap->start = best->end + 1;
			gap->end = filesize;
			filled->AddTail(gap);
		}
	}
	EMULE_CATCH
}

void CPartFile::UpdateFileRatingCommentAvail() {
	if (!this) return;

	bool prev=(hasComment || hasRating);

	hasComment=false;
	hasRating=false;

	POSITION pos1,pos2;
	for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
	for (pos1 = srclists[sl].GetHeadPosition();( pos2 = pos1 ) != NULL;){
		srclists[sl].GetNext(pos1);
		CUpDownClient* cur_src = srclists[sl].GetAt(pos2);
		if (cur_src->GetFileComment().GetLength()>0) hasComment=true;
		if (cur_src->GetFileRate()>0) hasRating=true;
		if (hasComment && hasRating) break;
	}
	if (prev!=(hasComment || hasRating)) UpdateDisplayedInfo();//theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
}

uint16 CPartFile::GetSourceCount() {
	uint16 count=0;
	for (int i=0;i<SOURCESSLOTS;i++) count+=srclists[i].GetCount();
	return count;
}

bool CPartFile::HasBadRating(){
	if (!hasRating) return false;

	POSITION pos1,pos2;
	for (int sl=0;sl<SOURCESSLOTS;sl++) if (!srclists[sl].IsEmpty())
	for (pos1 = srclists[sl].GetHeadPosition();( pos2 = pos1 ) != NULL;){
		srclists[sl].GetNext(pos1);
		CUpDownClient* cur_src = srclists[sl].GetAt(pos2);
		if (cur_src->GetFileRate()==1) return true;
	}
	return false;
}

void CPartFile::UpdateDisplayedInfo(boolean force) {
    DWORD curTick = ::GetTickCount();

    if(force || curTick-m_lastRefreshedDLDisplay > MINWAIT_BEFORE_DLDISPLAY_WINDOWUPDATE+(uint32)(rand()/(RAND_MAX/1000))) {
	    theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
        m_lastRefreshedDLDisplay = curTick;
    }
}
